# Generated by Django 4.2.18 on 2025-01-27 11:24

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0011_alter_datasource_user_editable_organizations"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="organization",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="organizationclass",
            unique_together=set(),
        ),
        migrations.AddConstraint(
            model_name="organization",
            constraint=models.UniqueConstraint(
                fields=("data_source", "origin_id"),
                name="django_orghierarchy_organization_data_source_and_origin_id_unique",
            ),
        ),
        migrations.AddConstraint(
            model_name="organization",
            constraint=models.UniqueConstraint(
                condition=models.Q(("data_source__isnull", True)),
                fields=("origin_id",),
                name="django_orghierarchy_organization_origin_id_unique_if_data_source_null",
            ),
        ),
        migrations.AddConstraint(
            model_name="organizationclass",
            constraint=models.UniqueConstraint(
                fields=("data_source", "origin_id"),
                name="django_orghierarchy_organizationclass_data_source_and_origin_id_unique",
            ),
        ),
        migrations.AddConstraint(
            model_name="organizationclass",
            constraint=models.UniqueConstraint(
                condition=models.Q(("data_source__isnull", True)),
                fields=("origin_id",),
                name="django_orghierarchy_organizationclass_origin_id_unique_if_data_source_null",
            ),
        ),
    ]



# Generated by Django 1.11.7 on 2017-12-13 18:42
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0002_organization_replaced_by"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="organization",
            name="responsible_organization",
        ),
        migrations.AddField(
            model_name="organization",
            name="internal_type",
            field=models.CharField(
                choices=[
                    ("normal", "Normal organization"),
                    ("affiliated", "Affiliated organization"),
                ],
                default="normal",
                max_length=20,
            ),
        ),
    ]



# Generated by Django 3.2 on 2023-02-02 08:18

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0010_rename_user_editable"),
    ]

    operations = [
        migrations.AlterField(
            model_name="datasource",
            name="user_editable_organizations",
            field=models.BooleanField(
                default=False, verbose_name="Organizations may be edited by users"
            ),
        ),
    ]



from unittest.mock import MagicMock, patch

from django.core.management import CommandError, call_command
from django.test import TestCase

from django_orghierarchy.importers import DataImportError
from django_orghierarchy.models import Organization, OrganizationClass
from django_orghierarchy.utils import get_data_source_model

from .test_importers import mock_request_get


class TestRestImportCommand(TestCase):
    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_rest_import_success(self):
        url = "http://fake.url/organizations/?page=1"
        call_command("import_organizations", url)

        data_source_model = get_data_source_model()
        self.assertEqual(Organization.objects.count(), 3)
        self.assertEqual(data_source_model.objects.count(), 2)
        self.assertEqual(OrganizationClass.objects.count(), 2)

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_rest_import_handle_data_import_error(self):
        url = "http://not-exist.url/organizatoins/?page=1"
        try:
            call_command("import_organizations", url)
        except DataImportError:
            self.fail("The command does not handle DataImportError")

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_rest_import_with_renamed_source(self):
        url = "http://fake.url/organizations/?page=1"
        rename_data_source = [
            "test-source-1: new-source-1",
            "test-source-2: new-source-2",
        ]
        call_command("import_organizations", url, rename_data_source=rename_data_source)

        data_source_model = get_data_source_model()
        qs = data_source_model.objects.filter(id__in=["new-source-1", "new-source-2"])
        self.assertTrue(qs.count(), 2)
        self.assertFalse(
            data_source_model.objects.filter(
                id__in=["test-source-1", "test-source-2"]
            ).exists()
        )

    def test_rest_import_with_invalid_renamed_source(self):
        url = "http://fake.url/organizations/?page=1"
        rename_data_source = [
            "test-source-1: new-source-1",
            "test-source-2, new-source-2",
        ]
        self.assertRaises(
            CommandError,
            call_command,
            "import_organizations",
            url,
            rename_data_source=rename_data_source,
        )



from rest_framework import serializers, viewsets

from .models import Organization, OrganizationClass
from .utils import get_data_source_model


class DataSourceSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_data_source_model()
        fields = "__all__"


class OrganizationClassSerializer(serializers.ModelSerializer):
    class Meta:
        model = OrganizationClass
        fields = "__all__"


class OrganizationSerializer(serializers.ModelSerializer):
    data_source = DataSourceSerializer()
    classification = OrganizationClassSerializer()

    class Meta:
        model = Organization
        fields = "__all__"


class OrganizationViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = Organization.objects.select_related("data_source", "classification")
    serializer_class = OrganizationSerializer



# Generated by Django 1.11.7 on 2018-01-09 20:02
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0004_add_affiliated_org_permissions"),
    ]

    operations = [
        migrations.AddField(
            model_name="organizationclass",
            name="created_time",
            field=models.DateTimeField(
                default=django.utils.timezone.now,
                help_text="The time at which the resource was created",
            ),
        ),
        migrations.AddField(
            model_name="organizationclass",
            name="data_source",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.DJANGO_ORGHIERARCHY_DATASOURCE_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="organizationclass",
            name="last_modified_time",
            field=models.DateTimeField(
                auto_now=True, help_text="The time at which the resource was updated"
            ),
        ),
        migrations.AddField(
            model_name="organizationclass",
            name="origin_id",
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AlterField(
            model_name="organization",
            name="created_time",
            field=models.DateTimeField(
                default=django.utils.timezone.now,
                help_text="The time at which the resource was created",
            ),
        ),
        migrations.AlterField(
            model_name="organization",
            name="origin_id",
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AlterField(
            model_name="organizationclass",
            name="id",
            field=models.CharField(
                editable=False, max_length=255, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="organizationclass",
            name="name",
            field=models.CharField(max_length=255),
        ),
        migrations.AlterUniqueTogether(
            name="organizationclass",
            unique_together=set([("data_source", "origin_id")]),
        ),
    ]



from unittest.mock import MagicMock, patch

from django.test import TestCase

from django_orghierarchy.models import DataSource
from django_orghierarchy.utils import get_data_source_model, swapper


class TestGetDataSourceModel(TestCase):
    def test_get_data_source_model_return_default_model(self):
        model = get_data_source_model()
        self.assertIs(model, DataSource)

    @patch("swapper.load_model", MagicMock())
    def test_get_data_source_model_swapper_load_model_called(self):
        get_data_source_model()
        self.assertTrue(swapper.load_model.called)



import datetime

import pytest
from pytest_django.asserts import assertQuerySetEqual

from django_orghierarchy.importers import DataImportError, RestAPIImporter
from django_orghierarchy.models import Organization, OrganizationClass
from django_orghierarchy.utils import get_data_source_model
from tests.factories import OrganizationFactory

req_mock = None


@pytest.fixture(autouse=True)
def global_requests_mock(requests_mock):
    global req_mock
    req_mock = requests_mock
    yield requests_mock

    req_mock = None


org_1 = {
    "abbreviation": "Hki",
    "classification": None,
    "created_at": "2014-11-27T20:06:22.404134",
    "deleted": False,
    "dissolution_date": None,
    "founding_date": None,
    "id": "hel:00001",
    "name_fi": "Helsingin kaupunki",
    "name_sv": "Helsingfors stad",
    "origin_id": "00001",
    "parents": [],
    "policymaker": None,
    "resource_uri": "/paatokset/v1/organization/hel%3A00001/",
    "slug": "00001",
    "type": "city",
    "updated_at": "2014-11-27T20:06:22.408446",
}

org_2 = {
    "abbreviation": "Kvsto",
    "classification": None,
    "created_at": "2014-11-27T20:06:22.618886",
    "deleted": False,
    "dissolution_date": None,
    "founding_date": None,
    "id": "hel:02900",
    "name_fi": "Kaupunginvaltuusto",
    "name_sv": "Stadsfullmäktige",
    "origin_id": "02900",
    "parents": ["/paatokset/v1/organization/hel%3A00001/"],
    "policymaker": "/paatokset/v1/policymaker/5/",
    "policymaker_slug": "kvsto",
    "resource_uri": "/paatokset/v1/organization/hel%3A02900/",
    "slug": "kvsto",
    "type": "council",
    "updated_at": "2017-08-03T11:42:34.949991",
}
org_3 = {
    "abbreviation": "Khs",
    "classification": None,
    "created_at": "2014-11-27T20:06:23.355894",
    "deleted": False,
    "dissolution_date": None,
    "founding_date": None,
    "id": "hel:00400",
    "name_fi": "Kaupunginhallitus",
    "name_sv": "Stadsstyrelsen",
    "origin_id": "00400",
    "parents": ["/paatokset/v1/organization/hel%3A02900/"],
    "policymaker": "/paatokset/v1/policymaker/2/",
    "policymaker_slug": "khs",
    "resource_uri": "/paatokset/v1/organization/hel%3A00400/",
    "slug": "khs",
    "type": "board",
    "updated_at": "2017-08-03T11:42:34.961049",
}
org_4 = {
    "abbreviation": "Keha",
    "classification": None,
    "created_at": "2017-08-03T11:42:42.491738",
    "deleted": False,
    "dissolution_date": None,
    "founding_date": "2017-06-01",
    "id": "hel:U50",
    "name_fi": "Keskushallinto",
    "name_sv": "Centralförvaltningen",
    "origin_id": "U50",
    "parents": ["/paatokset/v1/organization/hel%3A00400/"],
    "policymaker": None,
    "resource_uri": "/paatokset/v1/organization/hel%3AU50/",
    "slug": "u50",
    "type": "field",
    "updated_at": "2017-08-03T11:42:42.493469",
}
org_kanslia = {
    "abbreviation": "Kanslia",
    "classification": None,
    "created_at": "2017-08-03T11:42:42.566923",
    "deleted": False,
    "dissolution_date": None,
    "founding_date": "2018-01-15",
    "id": "hel:U02100",
    "name_fi": "Kaupunginkanslia",
    "name_sv": "Stadskansliet",
    "origin_id": "U02100",
    "parents": ["/paatokset/v1/organization/hel%3AU50/"],
    "policymaker": None,
    "resource_uri": "/paatokset/v1/organization/hel%3AU02100/",
    "slug": "u02100",
    "type": "packaged_service",
    "updated_at": "2018-02-13T14:06:14.671597",
}
org_talpa = {
    "abbreviation": "Talpa",
    "classification": None,
    "created_at": "2017-08-03T11:42:42.526192",
    "deleted": False,
    "dissolution_date": None,
    "founding_date": "2020-03-01",
    "id": "hel:U01600",
    "name_fi": "Taloushallintopalveluliikelaitos",
    "name_sv": "Affärsverket ekonomiförvaltningen",
    "origin_id": "U01600",
    "parents": ["/paatokset/v1/organization/hel%3AU50/"],
    "policymaker": None,
    "resource_uri": "/paatokset/v1/organization/hel%3AU01600/",
    "slug": "u01600",
    "type": "packaged_service",
    "updated_at": "2022-11-03T16:03:58.790299",
}


def register_responses():
    response_1 = {
        "meta": {
            "limit": 3,
            "next": "/organization/?limit=3&offset=3",
            "offset": 0,
            "previous": None,
            "total_count": 6,
        },
        "objects": [org_1, org_2, org_3],
    }
    response_2 = {
        "meta": {
            "limit": 3,
            "next": None,
            "offset": 3,
            "previous": "/organization/?limit=3&offset=0",
            "total_count": 6,
        },
        "objects": [org_4, org_kanslia, org_talpa],
    }

    req_mock.get("http://fake.url/organization/", json=response_1)
    req_mock.get("http://fake.url/organization/?limit=3&offset=3", json=response_2)


@pytest.fixture
def importer():
    register_responses()
    return RestAPIImporter(
        "http://fake.url/organization/", RestAPIImporter.openahjo_config
    )


def test_expected_importer_configuration(importer):
    assert importer.has_meta
    assert importer.next_key == "next"
    assert importer.results_key == "objects"
    assert importer.default_data_source == "OpenAhjoAPI"

    expected_field_config = {
        "classification": {"source_field": "type"},
        "name": {"source_field": "name_fi"},
        "origin_id": {"data_type": "str_lower"},
        "parent": {
            "data_type": "org_id_regex",
            "optional": True,
            "pattern": r"\/(\w+:\w+)\/$",
            "source_field": "parents",
            "unquote": True,
            "unwrap_list": True,
        },
    }
    assert importer.field_config == expected_field_config


def test_data_iter(importer):
    """Test that data_iter returns all items from the API."""
    iterator = importer._data_iter(importer.url)
    assert len(list(iterator)) == 6


@pytest.mark.parametrize(
    "resource_id,expected",
    [
        (
            "/organization/?limit=20&offset=20",
            "http://fake.url/organization/?limit=20&offset=20",
        ),
        ("123", "http://fake.url/organization/123"),
        (
            "https://identifier-with-host/paatokset/v1/organization/?limit=20",
            "https://identifier-with-host/paatokset/v1/organization/?limit=20",
        ),
    ],
)
def test_build_resource_url(importer, resource_id, expected):
    assert importer._build_resource_url(resource_id) == expected


@pytest.mark.django_db
def test_import_organization_class(importer):
    """Test that organization class is imported correctly.

    Currently supported types in Open Ahjo API:
    https://github.com/City-of-Helsinki/openahjo/blob/master/decisions/models.py#L11
    - council,
    - board,
    - board_division,
    - committee,
    - field,
    - department,
    - division,
    - introducer,
    - introducer_field,
    - office_holder,
    - trustee,
    - city,
    - unit
    """
    organization_class = importer._import_organization_class(org_1["type"])

    qs = OrganizationClass.objects.all()
    assertQuerySetEqual(qs, [organization_class])
    assert organization_class.id == f"{importer.default_data_source}:city"
    assert organization_class.name == f"{importer.default_data_source}:city"


@pytest.mark.django_db
def test_import_organization_class_with_remapped_data_source(
    importer,
):
    importer.config["rename_data_source"] = {
        f"{importer.default_data_source}": "remapped"
    }

    organization_class = importer._import_organization_class(org_1["type"])

    qs = OrganizationClass.objects.all()
    assertQuerySetEqual(qs, [organization_class])
    assert organization_class.id == "remapped:city"
    assert organization_class.name == "remapped:city"


@pytest.mark.django_db
def test_import_data_source(importer):
    """OpenAhjo API doesn't define a data source, so we use the default one."""
    data_source = importer._import_data_source(importer.default_data_source)

    data_source_model = get_data_source_model()
    qs = data_source_model.objects.all()
    assertQuerySetEqual(qs, [data_source])
    assert data_source.id == f"{importer.default_data_source}"


@pytest.mark.django_db
def test_import_data_source_with_remapped_identifier(importer):
    """Remapping the default data source."""
    importer.config["rename_data_source"] = {
        f"{importer.default_data_source}": "remapped"
    }

    data_source = importer._import_data_source(importer.default_data_source)

    data_source_model = get_data_source_model()
    qs = data_source_model.objects.all()
    assertQuerySetEqual(qs, [data_source])
    assert data_source.id == "remapped"


@pytest.mark.django_db
def test_import_data(importer):
    importer.import_data()
    data_source_model = get_data_source_model()
    assert Organization.objects.count() == 6
    assert data_source_model.objects.count() == 1
    assert OrganizationClass.objects.count() == 5


@pytest.mark.django_db
def test_import_organization_data(importer):
    importer.import_data()

    organization = Organization.objects.get(
        id=f"{importer.default_data_source}:{org_kanslia['origin_id'].lower()}"
    )
    assert organization.name == org_kanslia["name_fi"]
    assert (
        organization.classification.id
        == f"{importer.default_data_source}:packaged_service"
    )
    assert organization.data_source.id == f"{importer.default_data_source}"
    assert organization.origin_id == org_kanslia["origin_id"].lower()
    assert organization.founding_date == datetime.date(2018, 1, 15)
    assert organization.dissolution_date is None


@pytest.mark.django_db
def test_import_organization_data_with_remapped_data_source(importer):
    importer.config["rename_data_source"] = {
        f"{importer.default_data_source}": "remapped"
    }

    importer.import_data()

    organization = Organization.objects.get(
        id=f"remapped:{org_kanslia['origin_id'].lower()}"
    )
    assert organization.name == org_kanslia["name_fi"]
    assert organization.classification.id == "remapped:packaged_service"
    assert organization.data_source.id == "remapped"
    assert organization.origin_id == org_kanslia["origin_id"].lower()
    assert organization.founding_date == datetime.date(2018, 1, 15)
    assert organization.dissolution_date is None


@pytest.mark.django_db
def test_import_organization_with_parent(importer):
    importer.import_data()

    org_child = Organization.objects.get(
        id=f"{importer.default_data_source}:{org_kanslia['origin_id'].lower()}"
    )
    org_parent = Organization.objects.get(
        id=f"{importer.default_data_source}:{org_4['origin_id'].lower()}"
    )
    assert org_child.parent == org_parent


@pytest.mark.django_db
def test_import_organization_without_parent(importer):
    organization = importer._import_organization(org_1)

    qs = Organization.objects.all()
    assertQuerySetEqual(qs, [organization])
    assert organization.name == org_1["name_fi"]
    assert (
        organization.id
        == f"{importer.default_data_source}:{org_1['origin_id'].lower()}"
    )


@pytest.mark.django_db
def test_import_organization_flip_parents(importer):
    """Create organization with different parent organization than in the data and
    check that importing updates the parent organization."""
    organization_kanslia = OrganizationFactory(
        name="existing-organization",
        origin_id=org_kanslia["origin_id"].lower(),
        data_source=importer._import_data_source(importer.default_data_source),
    )

    organization_4 = OrganizationFactory(
        name="existing-organization",
        origin_id=org_4["origin_id"].lower(),
        data_source=importer._import_data_source(importer.default_data_source),
        parent=organization_kanslia,
    )

    importer.import_data()

    organization_kanslia.refresh_from_db()
    organization_4.refresh_from_db()

    assert organization_kanslia.parent == organization_4


@pytest.mark.django_db
def test_import_organization_update_existing(importer):
    organization = OrganizationFactory(
        name="existing-organization",
        origin_id=org_kanslia["origin_id"].lower(),
        data_source=importer._import_data_source(importer.default_data_source),
    )

    importer._import_organization(org_kanslia)

    organization.refresh_from_db()
    assert organization.name == org_kanslia["name_fi"]
    assert (
        organization.classification.id
        == f"{importer.default_data_source}:packaged_service"
    )
    assert organization.parent.name == org_4["name_fi"]
    assert organization.dissolution_date is None
    assert organization.founding_date == datetime.date(2018, 1, 15)


@pytest.mark.parametrize("skip_classifications", [True, False])
@pytest.mark.django_db
def test_import_organization_data_with_skip_classifications(
    importer, skip_classifications
):
    """Organization is not imported if it's in the classification skip list."""
    if skip_classifications:
        importer.config["skip_classifications"] = ["city"]

    importer.import_data()

    assert (
        Organization.objects.filter(origin_id__iendswith=org_1["origin_id"]).count()
        == 0
        if skip_classifications
        else 1
    )


@pytest.mark.parametrize(
    "value,unwrap,expected",
    [
        ([], True, None),
        (["identifier"], True, "identifier"),
        (["identifier1", "identifier2"], True, "identifier1"),
        ([], False, []),
        (["identifier"], False, ["identifier"]),
        (["identifier1", "identifier2"], False, ["identifier1", "identifier2"]),
    ],
)
def test_get_field_value_unwrap(importer, value, unwrap, expected):
    config = {"unwrap_list": unwrap}
    value = importer._get_field_value({"origin_id": value}, "origin_id", config)
    assert value == expected
    if expected is None:
        assert value is None


@pytest.mark.parametrize(
    "value,unquote,expected",
    [("a%3Ab%25c%23d", True, "a:b%c#d"), ("a%3Ab%25c%23d", False, "a%3Ab%25c%23d")],
)
def test_get_field_value_unquote(importer, value, unquote, expected):
    config = {
        "unquote": unquote,
    }
    value = importer._get_field_value({"origin_id": value}, "origin_id", config)
    assert value == expected


@pytest.mark.django_db
def test_get_field_value_org_id_regex_data_type(importer):
    """Parses organization id from a string using regex pattern."""
    config = {
        "source_field": "parents",
        "data_type": "org_id_regex",
        "unwrap_list": True,
        "unquote": True,
        "optional": True,
    }

    # test raise DataImportError if no pattern is provided
    with pytest.raises(DataImportError):
        importer._get_field_value(org_2, "parent", config)

    config["pattern"] = r"\/(\w+:\w+)\/$"

    # This should import the parent organization
    value = importer._get_field_value(org_2, "parent", config)
    assert isinstance(value, Organization)
    assert value.name == org_1["name_fi"]



import re
from unittest.mock import MagicMock, patch

import requests
from django.test import TestCase

from django_orghierarchy.importers import DataImportError, RestAPIImporter
from django_orghierarchy.models import Organization, OrganizationClass
from django_orghierarchy.utils import get_data_source_model

from .factories import OrganizationFactory


class MockResponse:
    def __init__(self, json_data, status_code=200):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data

    def raise_for_status(self):
        if self.status_code != 200:
            raise requests.HTTPError("HTTPError raised")


organization_1 = {
    "id": 111,
    "data_source": "test-source-1",
    "origin_id": "ABC-123",
    "classification": "test-source-1:test-class-1",
    "name": "Organization-1",
    "founding_date": "2000-01-01",
    "dissolution_date": "2017-01-01",
    "parent": "http://fake.url/organizations/222/",
    "ignored_field": "This field will be ignored",
}

organization_2 = {
    "id": 222,
    "data_source": "test-source-1",
    "origin_id": "ABC-456",
    "classification": "test-source-1:test-class-1",
    "name": "Organization-2",
    "founding_date": "2000-01-01",
    "dissolution_date": None,
    "parent": None,
    "ignored_field": "This field will be ignored",
}

organization_3 = {
    "id": 333,
    "data_source": "test-source-2",
    "origin_id": "XYZ-3",
    "classification": "test-source-2:test-class-2",
    "name": "Organizatoin-3",
    "founding_date": "2016-01-01",
    "dissolution_date": None,
    "parent": "http://fake.url/organizations/111/",
    "ignored_field": "This field will be ignored",
}

organizations = {
    "111": organization_1,
    "222": organization_2,
    "333": organization_3,
}

tprek_organization_1 = {
    "id": "111",
    "parent_id": "222",
    "organization_type": "TEST_TYPE_1",
    "name_fi": "Organization-1",
    "ignored_field": "This field will be ignored",
}

tprek_organization_2 = {
    "id": "222",
    "organization_type": "TEST_TYPE_1",
    "name_fi": "Organization-2",
    "ignored_field": "This field will be ignored",
}

# TPREK may have missing parent organizations. Pay no heed.
tprek_organization_3 = {
    "id": "333",
    "parent_id": "444",
    "organization_type": "TEST_TYPE_2",
    "name_fi": "Organization-3",
    "ignored_field": "This field will be ignored",
}

tprek_organizations = {
    "111": tprek_organization_1,
    "222": tprek_organization_2,
    "333": tprek_organization_3,
}


def mock_request_get(url, *args, **kwargs):
    m = re.search(r"http://fake.url/organizations/(\d+)/", url)
    if m:
        org_id = m.group(1)
        return MockResponse(organizations[org_id])

    m = re.search(r"http://fake.url/organizations/\?page=(\d+)$", url)
    if m:
        page = m.group(1)
        if page == "1":
            return MockResponse(
                {
                    "next": "http://fake.url/organizations/?page=2",
                    "results": [organization_1, organization_2],
                }
            )
        elif page == "2":
            return MockResponse({"next": None, "results": [organization_3]})
    m = re.search(r"http://fake.url/", url)
    if m:
        return MockResponse({"next_page": None, "items": []}, status_code=200)

    return MockResponse({}, status_code=404)


def mock_tprek_request_get(url, *args, **kwargs):
    m = re.search(r"http://fake.tprek.url/department/", url)
    if m:
        return MockResponse(tprek_organizations.values())

    return MockResponse({}, status_code=404)


class TestRestApiImporter(TestCase):
    def setUp(self):
        self.importer = self.get_importer()

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def get_importer(self):
        return RestAPIImporter("http://fake.url/organizations/?page=1")

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_custom_config(self):
        config = {
            "next_key": "next_page",
            "results_key": "items",
            "fields": ["classification", "origin_id", "name", "parent"],
            "field_config": {
                "classification": {
                    "data_type": "link",
                }
            },
        }
        importer = RestAPIImporter("http://fake.url/", config)
        self.assertEqual(importer.next_key, "next_page")
        self.assertEqual(importer.results_key, "items")

        expected_field_config = {
            "parent": {
                "data_type": "link",
            },
            "origin_id": {
                "data_type": "str_lower",
            },
            "classification": {
                "data_type": "link",
            },
        }
        self.assertDictEqual(importer.field_config, expected_field_config)

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_get_organization_class(self):
        data = {"id": "test-source:test-class"}
        organization_class = self.importer._get_organization_class(data)
        self.assertEqual(organization_class.id, "test-source:test-class")
        self.assertEqual(OrganizationClass.objects.count(), 1)

        self.importer._get_organization_class(data)
        self.assertEqual(OrganizationClass.objects.count(), 1)  # fetched from cached

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_get_data_source(self):
        data = {"id": "test-source"}
        data_source = self.importer._get_data_source(data)
        self.assertEqual(data_source.id, "test-source")
        data_source_model = get_data_source_model()
        self.assertEqual(data_source_model.objects.count(), 1)

        self.importer._get_data_source(data)
        self.assertEqual(data_source_model.objects.count(), 1)  # fetched from cached

        self.importer.config["rename_data_source"] = {"test-source": "new-source-name"}
        data_source = self.importer._get_data_source(data)
        self.assertEqual(data_source.id, "new-source-name")
        self.assertEqual(data_source_model.objects.count(), 2)

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_import_data(self):
        self.importer.import_data()
        data_source_model = get_data_source_model()
        self.assertEqual(Organization.objects.count(), 3)
        self.assertEqual(data_source_model.objects.count(), 2)
        self.assertEqual(OrganizationClass.objects.count(), 2)

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_import_organization_with_parent(self):
        organization = self.importer._import_organization(organization_1)
        qs = Organization.objects.all()
        # also created parent organization
        self.assertQuerySetEqual(qs, [organization.parent, organization], ordered=False)
        self.assertEqual(organization.name, "Organization-1")
        self.assertNotEqual(organization.id, 111)
        self.assertEqual(organization.parent.name, "Organization-2")
        self.assertNotEqual(organization.parent_id, 222)

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_import_organization_without_parent(self):
        organization = self.importer._import_organization(organization_2)
        qs = Organization.objects.all()
        self.assertQuerySetEqual(qs, [organization])
        self.assertEqual(organization.name, "Organization-2")
        self.assertNotEqual(organization.id, 222)

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_import_organization_flip_parents(self):
        organization = self.importer._import_organization(organization_1)
        qs = Organization.objects.all()
        # also created parent organization
        self.assertQuerySetEqual(qs, [organization.parent, organization], ordered=False)
        self.assertEqual(organization.name, "Organization-1")
        self.assertNotEqual(organization.id, 111)
        self.assertEqual(organization.parent.name, "Organization-2")
        self.assertNotEqual(organization.parent_id, 222)
        changed_organization = organization_2.copy()
        new_parent = organization_1.copy()
        changed_organization["parent"] = "http://fake.url/organizations/111/"
        new_parent["parent"] = None
        organizations["111"] = new_parent

        # Re-initialize the importer to clear the caches and data
        importer = self.get_importer()
        organization = importer._import_organization(changed_organization)
        # Now the parents should have switched.
        self.assertQuerySetEqual(qs, [organization.parent, organization], ordered=False)
        self.assertEqual(organization.name, "Organization-2")
        self.assertNotEqual(organization.id, 222)
        self.assertEqual(organization.parent.name, "Organization-1")
        self.assertNotEqual(organization.id, 111)
        # tear down this test, as it had a side effect
        organizations["111"] = organization_1

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_import_organization_update_existing(self):
        organization = OrganizationFactory(
            name="existing-organization",
            origin_id=organization_2["origin_id"],
            data_source=self.importer._import_data_source(
                organization_2["data_source"]
            ),
        )
        self.importer._import_organization(organization_2)
        organization.refresh_from_db()

        self.assertQuerySetEqual(Organization.objects.all(), [organization])
        self.assertEqual(organization.name, "Organization-2")

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_import_organization_data_with_skip_classifications(self):
        """Organization is not imported if it's in the classification skip list."""
        self.importer.config["skip_classifications"] = ["test-class-1"]
        organization = self.importer._import_organization(organization_1)
        self.assertIsNone(organization)

        # Re-initialize the importer to clear the caches and data
        importer = self.get_importer()
        organization = importer._import_organization(organization_1)
        self.assertEqual(organization.id, "test-source-1:abc-123")

    def test_import_data_source_with_string(self):
        data_source = self.importer._import_data_source("test-data-source")
        data_source_model = get_data_source_model()
        qs = data_source_model.objects.all()
        self.assertQuerySetEqual(qs, [data_source])
        self.assertEqual(data_source.id, "test-data-source")

    def test_import_data_source_with_dict_data(self):
        data = {"id": "test-data-source"}
        data_source = self.importer._import_data_source(data)
        data_source_model = get_data_source_model()
        qs = data_source_model.objects.all()
        self.assertQuerySetEqual(qs, [data_source])
        self.assertEqual(data_source.id, "test-data-source")

    def test_import_organization_class_with_string(self):
        organization_class = self.importer._import_organization_class(
            "test-source-1:test-org-class"
        )
        qs = OrganizationClass.objects.all()
        self.assertQuerySetEqual(qs, [organization_class])
        self.assertEqual(organization_class.id, "test-source-1:test-org-class")
        self.assertEqual(organization_class.name, "test-source-1:test-org-class")

    def test_import_organization_class_with_simple_string(self):
        organization_class = self.importer._import_organization_class(
            "class-with-no-source"
        )
        qs = OrganizationClass.objects.all()
        self.assertQuerySetEqual(qs, [organization_class])
        self.assertEqual(organization_class.id, "OpenDecisionAPI:class-with-no-source")
        self.assertEqual(
            organization_class.name, "OpenDecisionAPI:class-with-no-source"
        )

    def test_import_organization_class_with_simple_string_and_remapped_data_source(
        self,
    ):
        self.importer.config["rename_data_source"] = {"OpenDecisionAPI": "remapped"}

        organization_class = self.importer._import_organization_class(
            "class-with-no-source"
        )

        qs = OrganizationClass.objects.all()
        self.assertQuerySetEqual(qs, [organization_class])
        self.assertEqual(organization_class.id, "remapped:class-with-no-source")
        self.assertEqual(organization_class.name, "remapped:class-with-no-source")

    def test_import_organization_class_with_dict_data(self):
        data = {
            "id": 999,
            "origin_id": "test-org-class",
            "name": "test-org-class",
            "data_source": "test-source-1",
        }
        organization_class = self.importer._import_organization_class(data)
        qs = OrganizationClass.objects.all()
        self.assertQuerySetEqual(qs, [organization_class])
        self.assertEqual(organization_class.name, "test-org-class")
        self.assertEqual(organization_class.id, "test-source-1:test-org-class")

    def test_import_organization_class_with_simple_dict_data(self):
        data = {"id": 999, "name": "test-org-class"}
        organization_class = self.importer._import_organization_class(data)
        qs = OrganizationClass.objects.all()
        self.assertQuerySetEqual(qs, [organization_class])
        self.assertEqual(organization_class.name, "test-org-class")
        self.assertEqual(organization_class.id, "OpenDecisionAPI:999")

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_data_iter(self):
        url = "http://fake.url/organizations/?page=1"
        iterator = self.importer._data_iter(url)
        iter_len = len(list(iterator))
        self.assertEqual(iter_len, 3)

        url = "http://fake.url/organizations/?page=2"
        iterator = self.importer._data_iter(url)
        iter_len = len(list(iterator))
        self.assertEqual(iter_len, 1)

        url = "http://not-exist.url/organizations/"
        self.assertRaises(DataImportError, list, self.importer._data_iter(url))

    def test_get_field_value_empty_value(self):
        value = self.importer._get_field_value(organization_2, "dissolution_date", {})
        self.assertIsNone(value)

    def test_get_field_value_different_source_field(self):
        config = {
            "source_field": "id",
        }
        value = self.importer._get_field_value(organization_1, "origin_id", config)
        self.assertEqual(value, 111)

    def test_get_field_value_not_exist_field(self):
        config = {
            "source_field": "not-exist-field",
        }
        self.assertRaises(
            DataImportError,
            self.importer._get_field_value,
            organization_1,
            "origin_id",
            config,
        )

    def test_get_field_value_not_exist_data_type(self):
        config = {
            "data_type": "not-exist-data-type",
        }
        self.assertRaises(
            DataImportError,
            self.importer._get_field_value,
            organization_1,
            "origin_id",
            config,
        )

    @patch("requests.get", MagicMock(return_value=MockResponse("test-get-return")))
    def test_get_field_value_link_data_type(self):
        config = {
            "data_type": "link",
        }
        value = self.importer._get_field_value(
            {"name": "http://fake.url/"}, "name", config
        )
        self.assertEqual(value, "test-get-return")

    def test_get_field_value_str_lower_data_type(self):
        config = {
            "data_type": "str_lower",
        }
        value = self.importer._get_field_value(organization_1, "origin_id", config)
        self.assertEqual(value, "abc-123")

    def test_get_field_value_regex_data_type(self):
        config = {
            "data_type": "regex",
        }
        # test raise DataImportError if no pattern is provided
        self.assertRaises(
            DataImportError,
            self.importer._get_field_value,
            organization_1,
            "origin_id",
            config,
        )

        config = {
            "data_type": "regex",
            "pattern": r"\w+\-(\d+)",
        }
        value = self.importer._get_field_value(organization_1, "origin_id", config)
        self.assertEqual(value, "123")

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_get_field_value_related_fields(self):
        value = self.importer._get_field_value(organization_1, "data_source", {})
        self.assertEqual(value.id, "test-source-1")

        value = self.importer._get_field_value(organization_1, "classification", {})
        self.assertEqual(value.id, "test-source-1:test-class-1")

        value = self.importer._get_field_value(
            organization_1, "parent", {"data_type": "link"}
        )
        self.assertEqual(value.name, "Organization-2")

    def test_get_regex_data(self):
        pattern = r"\w+\-(\d+)"
        self.assertRaises(
            DataImportError,
            self.importer._get_regex_data,
            "123",
            pattern,
        )

        data = self.importer._get_regex_data("abc-123", pattern)
        self.assertEqual(data, "123")

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_get_link_data(self):
        invalid_url = "abc.123"
        self.assertRaises(DataImportError, self.importer._get_link_data, invalid_url)

        url = "http://not-exist.url/organizations/111/"
        self.assertRaises(DataImportError, self.importer._get_link_data, url)

        url = "http://fake.url/organizations/111/"
        data = self.importer._get_link_data(url)
        self.assertEqual(data, organization_1)


class TestTprekRestApiImporter(TestRestApiImporter):
    # here we test all the features of the TPREK import that are different from default REST import

    def setUp(self):
        self.importer = self.get_importer()

    @patch("requests.get", MagicMock(side_effect=mock_tprek_request_get))
    def get_importer(self):
        config = {
            "next_key": None,
            "results_key": None,
            "fields": [
                "origin_id",
                "classification",
                "name",
                "parent",
            ],
            "update_fields": [
                "classification",
                "name",
                "parent",
            ],
            "field_config": {
                "parent": {
                    "source_field": "parent_id",
                    "data_type": "org_id",
                    "optional": True,
                },
                "origin_id": {
                    "source_field": "id",
                    "data_type": "value",
                },
                "classification": {
                    "source_field": "organization_type",
                    "optional": True,
                },
                "name": {
                    "source_field": "name_fi",
                    "optional": True,
                },
            },
            "default_data_source": "tprek",
            "default_parent_organization": "Pääkaupunkiseudun toimipisterekisteri",
        }
        return RestAPIImporter("http://fake.tprek.url/department/", config)

    @patch("requests.get", MagicMock(side_effect=mock_tprek_request_get))
    def test_get_data_source(self):
        data = {"id": "tprek"}
        data_source = self.importer._get_data_source(data)
        self.assertEqual(data_source.id, "tprek")
        data_source_model = get_data_source_model()
        self.assertEqual(data_source_model.objects.count(), 1)

        self.importer._get_data_source(data)
        self.assertEqual(data_source_model.objects.count(), 1)  # fetched from cached

        self.importer.config["rename_data_source"] = {"tprek": "new-source-name"}
        data_source = self.importer._get_data_source(data)
        self.assertEqual(data_source.id, "new-source-name")
        self.assertEqual(data_source_model.objects.count(), 2)

    @patch("requests.get", MagicMock(side_effect=mock_tprek_request_get))
    def test_get_field_value_related_fields(self):
        value = self.importer._get_field_value(
            {"data_source": "tprek"}, "data_source", {}
        )
        self.assertEqual(value.id, "tprek")

        value = self.importer._get_field_value(
            tprek_organization_1,
            "classification",
            self.importer.config["field_config"]["classification"],
        )
        self.assertEqual(value.id, "tprek:TEST_TYPE_1")

        value = self.importer._get_field_value(
            tprek_organization_1,
            "parent",
            self.importer.config["field_config"]["parent"],
        )
        self.assertEqual(value.id, "tprek:222")

    @patch("requests.get", MagicMock(side_effect=mock_tprek_request_get))
    def test_data_iter(self):
        url = "http://fake.tprek.url/department/"
        iterator = self.importer._data_iter(url)
        iter_len = len(list(iterator))
        self.assertEqual(iter_len, 3)

        url = "http://not-exist.url/organizations/"
        self.assertRaises(DataImportError, list, self.importer._data_iter(url))

    @patch("requests.get", MagicMock(side_effect=mock_tprek_request_get))
    def test_import_data(self):
        self.importer.import_data()
        data_source_model = get_data_source_model()
        self.assertEqual(Organization.objects.count(), 4)
        self.assertEqual(data_source_model.objects.count(), 1)
        self.assertEqual(OrganizationClass.objects.count(), 2)

    def test_import_organization_class_with_simple_string(self):
        organization_class = self.importer._import_organization_class(
            "class-with-no-source"
        )
        qs = OrganizationClass.objects.all()
        self.assertQuerySetEqual(qs, [organization_class])
        self.assertEqual(organization_class.id, "tprek:class-with-no-source")
        self.assertEqual(organization_class.name, "tprek:class-with-no-source")

    def test_import_organization_class_with_simple_string_and_remapped_data_source(
        self,
    ):
        self.importer.config["rename_data_source"] = {"tprek": "remapped"}

        organization_class = self.importer._import_organization_class(
            "class-with-no-source"
        )

        qs = OrganizationClass.objects.all()
        self.assertQuerySetEqual(qs, [organization_class])
        self.assertEqual(organization_class.id, "remapped:class-with-no-source")
        self.assertEqual(organization_class.name, "remapped:class-with-no-source")

    def test_import_organization_class_with_dict_data(self):
        pass

    def test_import_organization_class_with_simple_dict_data(self):
        pass

    def test_import_data_source_with_string(self):
        pass

    def test_import_data_source_with_dict_data(self):
        pass

    @patch("requests.get", MagicMock(side_effect=mock_tprek_request_get))
    def test_import_organization_data_with_skip_classifications(self):
        """Organization is not imported if it's in the classification skip list."""
        self.importer.config["skip_classifications"] = ["TEST_TYPE_1"]
        organization = self.importer._import_organization(tprek_organization_1)
        self.assertIsNone(organization)

        # Re-initialize the importer to clear the caches and data
        importer = self.get_importer()
        organization = importer._import_organization(tprek_organization_1)
        self.assertEqual(organization.id, "tprek:111")

    def test_import_organization_update_existing(self):
        organization = OrganizationFactory(
            name="existing-organization",
            origin_id=tprek_organization_2["id"],
            data_source=self.importer._import_data_source("tprek"),
        )
        self.importer._import_organization(tprek_organization_2)
        organization.refresh_from_db()
        default_parent = Organization.objects.get(id="tprek:tprek")

        self.assertQuerySetEqual(
            Organization.objects.all(),
            [default_parent, organization],
            ordered=False,
        )
        self.assertEqual(organization.name, "Organization-2")

    @patch("requests.get", MagicMock(side_effect=mock_tprek_request_get))
    def test_import_organization_with_parent(self):
        organization = self.importer._import_organization(tprek_organization_1)
        qs = Organization.objects.all()
        default_parent = Organization.objects.get(id="tprek:tprek")

        # also created parent organization.
        # parent will have the default parent organization
        self.assertQuerySetEqual(
            qs,
            [default_parent, organization.parent, organization],
            ordered=False,
        )
        self.assertEqual(organization.name, "Organization-1")
        self.assertEqual(organization.id, "tprek:111")
        self.assertEqual(organization.parent.name, "Organization-2")
        self.assertEqual(organization.parent_id, "tprek:222")
        self.assertEqual(
            organization.parent.parent.name, "Pääkaupunkiseudun toimipisterekisteri"
        )
        self.assertEqual(organization.parent.parent_id, "tprek:tprek")

    @patch("requests.get", MagicMock(side_effect=mock_request_get))
    def test_import_organization_with_missing_parent(self):
        organization = self.importer._import_organization(tprek_organization_3)
        qs = Organization.objects.all()

        # also created parent organization.
        # parent was not found, so organization will have the default
        default_parent = Organization.objects.get(id="tprek:tprek")
        self.assertQuerySetEqual(qs, [default_parent, organization])
        self.assertEqual(organization.name, "Organization-3")
        self.assertEqual(organization.id, "tprek:333")
        self.assertEqual(
            organization.parent.name, "Pääkaupunkiseudun toimipisterekisteri"
        )
        self.assertEqual(organization.parent_id, "tprek:tprek")

    @patch("requests.get", MagicMock(side_effect=mock_tprek_request_get))
    def test_import_organization_without_parent(self):
        organization = self.importer._import_organization(tprek_organization_2)
        qs = Organization.objects.all()
        default_parent = Organization.objects.get(id="tprek:tprek")
        self.assertQuerySetEqual(qs, [default_parent, organization], ordered=False)
        self.assertEqual(organization.name, "Organization-2")
        self.assertEqual(organization.id, "tprek:222")
        # organization will have the default parent organization
        self.assertEqual(
            organization.parent.name, "Pääkaupunkiseudun toimipisterekisteri"
        )
        self.assertEqual(organization.parent_id, "tprek:tprek")

    @patch("requests.get", MagicMock(side_effect=mock_tprek_request_get))
    def test_import_organization_flip_parents(self):
        organization = self.importer._import_organization(tprek_organization_1)
        qs = Organization.objects.all()
        default_parent = Organization.objects.get(id="tprek:tprek")

        # also created parent organization.
        # parent will have the default parent organization
        self.assertQuerySetEqual(
            qs,
            [default_parent, organization.parent, organization],
            ordered=False,
        )
        self.assertEqual(organization.name, "Organization-1")
        self.assertEqual(organization.id, "tprek:111")
        self.assertEqual(organization.parent.name, "Organization-2")
        self.assertEqual(organization.parent_id, "tprek:222")
        self.assertEqual(
            organization.parent.parent.name, "Pääkaupunkiseudun toimipisterekisteri"
        )
        self.assertEqual(organization.parent.parent_id, "tprek:tprek")

        changed_organization = tprek_organization_2.copy()
        new_parent = tprek_organization_1.copy()
        changed_organization["parent_id"] = "111"
        new_parent["parent_id"] = None
        tprek_organizations["111"] = new_parent

        # We must init a new importer, since importing by organization id will use a cached dict of all the
        # organizations in the importer.
        new_importer = self.get_importer()
        organization = new_importer._import_organization(changed_organization)
        # Now the parents should have switched.
        self.assertQuerySetEqual(
            qs,
            [default_parent, organization.parent, organization],
            ordered=False,
        )
        self.assertEqual(organization.name, "Organization-2")
        self.assertEqual(organization.id, "tprek:222")
        self.assertEqual(organization.parent.name, "Organization-1")
        self.assertEqual(organization.parent_id, "tprek:111")
        self.assertEqual(
            organization.parent.parent.name, "Pääkaupunkiseudun toimipisterekisteri"
        )
        self.assertEqual(organization.parent.parent_id, "tprek:tprek")
        # tear down this test, as it had a side effect
        tprek_organizations["111"] = tprek_organization_1



from functools import reduce

import swapper
from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from mptt.admin import DraggableMPTTAdmin

from .forms import AffiliatedOrganizationForm, OrganizationForm, SubOrganizationForm
from .models import Organization, OrganizationClass
from .utils import get_data_source_model

data_source_model = swapper.get_model_name("django_orghierarchy", "DataSource")
# Only register admin when using default data source model
# When the data source model is swapped, client code should
# be responsible for creating admin page for the model
if data_source_model == "django_orghierarchy.DataSource":

    @admin.register(get_data_source_model())
    class DataSourceAdmin(admin.ModelAdmin):
        list_display = ("id", "name", "user_editable_organizations")


@admin.register(OrganizationClass)
class OrganizationClassAdmin(admin.ModelAdmin):
    list_display = ("id", "name")
    list_filter = ("data_source",)


class SubOrganizationInline(admin.TabularInline):
    model = Organization
    verbose_name = _("added sub organization")
    verbose_name_plural = _("added sub organizations")
    fk_name = "parent"
    form = SubOrganizationForm
    organization_type = Organization.NORMAL
    # these have to be specified separately to prevent seeing redundant internal_type
    # field
    fields = ("name", "founding_date", "classification", "data_source", "origin_id")
    extra = 0

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        return queryset.filter(
            internal_type=self.organization_type,
            data_source__user_editable_organizations=True,
        )

    def has_add_permission(self, request, obj=None):
        return False

    def get_readonly_fields(self, request, obj=None):
        return ("data_source", "origin_id", "id")


class AddSubOrganizationInline(admin.TabularInline):
    model = Organization
    verbose_name = _("add sub organization")
    verbose_name_plural = _("add sub organizations")
    fk_name = "parent"
    form = SubOrganizationForm
    organization_type = Organization.NORMAL
    # these have to be specified separately to prevent seeing redundant internal_type
    # field
    fields = ("name", "founding_date", "classification", "data_source", "origin_id")
    extra = 1

    def get_queryset(self, request):
        return super().get_queryset(request).none()

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_readonly_fields(self, request, obj=None):
        # we have to have this separate since obj is the parent org, not the new empty
        # one so new suborganizations to be added allow all fields to be defined
        return ("id",)


class ProtectedSubOrganizationInline(admin.TabularInline):
    model = Organization
    verbose_name = _("non-editable sub organization")
    verbose_name_plural = _("non-editable sub organizations")
    fk_name = "parent"
    form = SubOrganizationForm
    organization_type = Organization.NORMAL
    # these have to be specified separately to prevent seeing redundant internal_type
    # field
    fields = ("name", "founding_date", "classification", "data_source", "origin_id")
    extra = 0

    def get_readonly_fields(self, request, obj=None):
        return self.form.base_fields

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        return queryset.filter(
            internal_type=self.organization_type,
            data_source__user_editable_organizations=False,
        )

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        # here obj refers to the *parent* organization, change permission to parent is
        # needed
        if request.user.has_perm("django_orghierarchy.change__organization"):
            return True
        return super().has_change_permission(request, obj)


class AffiliatedOrganizationInline(SubOrganizationInline):
    verbose_name = _("added affiliated organization")
    verbose_name_plural = _("added affiliated organizations")
    form = AffiliatedOrganizationForm
    organization_type = Organization.AFFILIATED

    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        if request.user.has_perm(
            "django_orghierarchy.change__organization"
        ) or request.user.has_perm(
            "django_orghierarchy.change_affiliated_organization"
        ):
            return True
        return super().has_change_permission(request, obj)

    def has_delete_permission(self, request, obj=None):
        # has_change_permission must be True to allow listing, even in read only
        if request.user.has_perm(
            "django_orghierarchy.delete__organization"
        ) or request.user.has_perm(
            "django_orghierarchy.delete_affiliated_organization"
        ):
            return True
        return super().has_delete_permission(request, obj)


class AddAffiliatedOrganizationInline(AddSubOrganizationInline):
    verbose_name = _("add affiliated organization")
    verbose_name_plural = _("add affiliated organizations")
    form = AffiliatedOrganizationForm
    organization_type = Organization.AFFILIATED

    def __init__(self, *args, **kwargs):
        # this is the only admin that needs the (hidden) internal_type field, to create
        # a new affiliated organization
        super().__init__(*args, **kwargs)
        self.fields += ("internal_type",)

    def has_add_permission(self, request, obj=None):
        if request.user.has_perm(
            "django_orghierarchy.add__organization"
        ) or request.user.has_perm("django_orghierarchy.add_affiliated_organization"):
            return True
        return super().has_add_permission(request, obj)


class ProtectedAffiliatedOrganizationInline(ProtectedSubOrganizationInline):
    verbose_name = _("non-editable affiliated organization")
    verbose_name_plural = _("non-editable affiliated organizations")
    form = AffiliatedOrganizationForm
    organization_type = Organization.AFFILIATED
    extra = 0

    def has_change_permission(self, request, obj=None):
        # here obj refers to the *parent* organization, change permission to parent is
        # needed
        if request.user.has_perm(
            "django_orghierarchy.change__organization"
        ) or request.user.has_perm(
            "django_orghierarchy.change_affiliated_organization"
        ):
            return True
        return super().has_change_permission(request, obj)


@admin.register(Organization)
class OrganizationAdmin(DraggableMPTTAdmin):
    filter_horizontal = ("admin_users", "regular_users")
    form = OrganizationForm
    inlines = [
        ProtectedSubOrganizationInline,
        SubOrganizationInline,
        AddSubOrganizationInline,
        ProtectedAffiliatedOrganizationInline,
        AffiliatedOrganizationInline,
        AddAffiliatedOrganizationInline,
    ]
    search_fields = ("name", "id", "origin_id", "classification__id")
    list_display = (*DraggableMPTTAdmin.list_display, "identifier", "classification_id")
    list_filter = ("data_source",)

    # these fields may not be changed at all in existing organizations
    existing_readonly_fields = ("id", "data_source", "origin_id", "internal_type")
    # these fields may not be changed at all in protected organizations
    protected_readonly_fields = existing_readonly_fields + (
        "origin_id",
        "classification",
        "name",
        "founding_date",
        "dissolution_date",
        "parent",
    )

    def identifier(self, obj):
        # Disable sorting on this column
        return obj.id

    def get_queryset(self, request):
        if not request.user.is_superuser:
            if not request.user.admin_organizations.all():
                return Organization.objects.none()
            # regular admins have rights to all organizations below their level
            admin_orgs = []
            for admin_org in request.user.admin_organizations.all():
                admin_orgs.append(admin_org.get_descendants(include_self=True))
            # for multiple admin_orgs, we have to combine the querysets and filter
            # distinct
            return reduce(lambda a, b: a | b, admin_orgs).distinct()
        return super().get_queryset(request)

    def get_readonly_fields(self, request, obj=None):
        has_write_access = False
        # queryset is already filtered, but write permissions have to be checked based
        # on organization_type
        if obj and obj.internal_type == Organization.AFFILIATED:
            # full rights also cover affiliated organizations
            has_write_access = request.user.has_perm(
                "django_orghierarchy.change_organization"
            ) or request.user.has_perm(
                "django_orghierarchy.change_affiliated_organization"
            )
        else:
            has_write_access = request.user.has_perm(
                "django_orghierarchy.change_organization"
            )
        if obj and not has_write_access:
            # has_change_permission has been evaluated to True (i.e. user can
            # open change form in admin) if user has change permissions on
            # affiliated organizations, this i  s to make sure user cannot edit
            # current organization if he does not have change permission on
            # organization
            readonly_fields = self.form.base_fields.copy()
            if request.user.has_perm(
                "django_orghierarchy.change_organization_regular_users"
            ):
                del readonly_fields["regular_users"]
            return readonly_fields
        if obj and obj.data_source and not obj.data_source.user_editable_organizations:
            # Organization data from protected data sources may not be edited
            if not request.user.has_perm("django_orghierarchy.replace_organization"):
                # Replacing an organization in the hierarchy requires extra privileges
                return self.protected_readonly_fields + ("replaced_by",)
            # Protected organizations allow only user fields to be changed
            return self.protected_readonly_fields
        # Editable organizations have no such restrictions, but replacement still
        # requires extra privileges
        readonly_fields = super().get_readonly_fields(request, obj)
        if obj:
            readonly_fields += self.existing_readonly_fields
        if not request.user.has_perm("django_orghierarchy.replace_organization"):
            return readonly_fields + ("replaced_by",)
        return readonly_fields

    def save_model(self, request, obj, form, change):
        # TODO: only allow creating objects for editable data sources?! or only system data source? # noqa E501
        if not obj.pk:
            obj.created_by = request.user
        obj.last_modified_by = request.user
        obj.save()

    def indented_title(self, item):
        """
        Override base method to add custom styles for affiliated organizations
        """
        additional_styles = ""
        if item.internal_type == Organization.AFFILIATED:
            additional_styles = "color: red;"

        return format_html(
            '<div style="text-indent:{}px; {}">{}</div>',
            item._mpttfield("level") * self.mptt_level_indent,
            additional_styles,
            item,
        )

    def has_change_permission(self, request, obj=None):
        if request.user.has_perm(
            "django_orghierarchy.change_affiliated_organization"
        ) or request.user.has_perm(
            "django_orghierarchy.change_organization_regular_users"
        ):
            # allow changing affiliated organization means user can also
            # changing current organization (open change form)
            return True
        return super().has_change_permission(request, obj)

    def get_actions(self, request):
        actions = super().get_actions(request)

        if (
            not request.user.has_perm("django_orghierarchy.delete_organization")
            and "delete_selected" in actions
        ):
            del actions["delete_selected"]
        return actions



from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from .factories import DataSourceFactory, OrganizationFactory


class OrganizationAPITestCase(APITestCase):
    def setUp(self):
        data_source = DataSourceFactory(name="abc")
        self.organization = OrganizationFactory(
            data_source=data_source, origin_id="123"
        )

    def test_organization_list(self):
        url = reverse("api:organization-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_organization(self):
        url = reverse("api:organization-detail", kwargs={"pk": self.organization.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_modify_organization(self):
        url = reverse("api:organization-detail", kwargs={"pk": self.organization.pk})
        response = self.client.post(url, {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_create_organization(self):
        url = reverse("api:organization-list")
        response = self.client.post(url, {})
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)



import copy
import logging
import re
import urllib.parse
import warnings
from enum import Enum

import requests
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.db import transaction

from .models import Organization, OrganizationClass
from .utils import get_data_source_model

logger = logging.getLogger(__name__)


class DataImportError(Exception):
    pass


class DataType(Enum):
    VALUE = "value"
    STR_LOWER = "str_lower"
    LINK = "link"
    REGEX = "regex"
    ORG_ID = "org_id"
    ORG_ID_REGEX = "org_id_regex"


class RestAPIImporter:
    r"""This class allows importing organization data from a REST API endpoint. The
    default configuration supports the 6aika Open Decision API specification:
    https://github.com/6aika/api-paatos

    The importer will also create data sources and organization classes if does not
    exist. If a single value is provided to `data_source` and `classification` fields,
    then the importer assumes the value is the name of the data source or organization
    class. If an object (dict) is provided to `data_source` and `classification` fields,
    then the importer will create objects with object data. Either full object data or a
    link to full object data should be provided to parent field so that the importer can
    create parent organization before creating child organization.

    For use in multi-source setups, the importer may optionally create a parent
    organization for the data source that is the common parent of all imported top-level
    organizations. This way, separate organization hierarchies from different data
    sources may be stored in the same application.

    The client code can override the default config or pass in a config object to allow
    the importer parsing different REST API structures.

    Config options:
        - next_key: The url to the next page if the source data is paginated, or None if
            not.
        - results_key: The object key to the list of organization objects, or None if
            the list is at root level.
        - fields: The fields of which the values will be imported. The same fields will
            be imported in organization classes, if present.
        - update_fields: The fields to update if the organization with same origin_id
            and data source exists.
        - skip_classifications: List of organization classifications which will cause
            organizations of this type to be skipped upon import.
        - has_meta: Does the JSON contain a separate object for pagination. Default
            False.
        - deprecated: Is the configuration considered deprecated. Default False.
        - field_config: Configs for each field. Config options:
            - source_field: The source data object key where the field value comes from.
                Defaults to original field name.
            - data_type: The data type for the field, can be 'value', 'str_lower',
                'link', 'regex', 'org_id' or 'org_id_regex'.
                Defaults to 'value'. If the data type is 'value', it will return the
                value from source data; if the data type is 'str_lower', it will return
                the value stringified and lowercased; if the data type is 'link', it
                will return the data fetched from the link; if the data type is 'regex',
                it will return the value extracted from the given pattern. If the data
                type is 'org_id' or 'org_id_regex', it will return the organization in
                the JSON with the given 'id'.
            - pattern: The pattern to extract the value from the source data. Only used
                when data_type is 'regex' or 'org_id_regex'.
            - unquote: Should the value be run through urllib.parse.unquote. Defaults to
                 False.
            - unwrap_list: Value is wrapped inside of a list. Return the first value on
                the list. Defaults to False.
            - optional: Whether the importer will continue if the field is missing.
                Defaults to False.
        - rename_data_source: Data sources that are renamed during import.
        - default_data_source: Data source id to use for objects without data source.
        - default_parent_organization: If set, name of the parent organization to be
            created for all imported top level organizations.

    Example config:
        {
            "next_key": "next",
            "results_key": "results",
            "fields": ["origin_id", "data_source", "classification"],
            "update_fields": ["classification"],
            "field_config": {
                "origin_id": {
                    "source_field": "id",
                    "data_type": "regex",
                    "pattern": r"abc:(\d+)",
                },
                "parent": {
                    "data_type": "link",
                },
            },
            "rename_data_source": {
                "original_name_1": "new_name_1",
                "original_name_2": "new_name_2",
            },
            "default_data_source": "new_name_1",
            "default_parent_organization": None,
        }
    """

    # https://api.hel.fi/paatos/v1/organization/
    paatos_config = {
        "deprecated": True,
        "has_meta": False,
        "next_key": "next",
        "results_key": "results",
        "fields": [
            "data_source",
            "origin_id",
            "classification",
            "name",
            "founding_date",
            "dissolution_date",
            "parent",
        ],
        "update_fields": [
            "classification",
            "name",
            "founding_date",
            "dissolution_date",
            "parent",
        ],
        "field_config": {
            "parent": {
                "data_type": "link",
            },
            "origin_id": {
                "data_type": "str_lower",
            },
        },
        "default_data_source": "OpenDecisionAPI",
    }

    # https://dev.hel.fi/apis/openahjo
    # https://dev.hel.fi/paatokset/v1/organization/
    # https://github.com/City-of-Helsinki/openahjo
    openahjo_config = {
        "has_meta": True,
        "next_key": "next",
        "results_key": "objects",
        "fields": [
            "origin_id",
            "classification",
            "name",
            "founding_date",
            "dissolution_date",
            "parent",
        ],
        "skip_classifications": ["office_holder", "team", "subteam", "trustee"],
        "update_fields": [
            "classification",
            "name",
            "founding_date",
            "dissolution_date",
            "parent",
        ],
        "field_config": {
            "name": {
                "source_field": "name_fi",
            },
            "classification": {
                "source_field": "type",
            },
            "parent": {
                "source_field": "parents",
                "data_type": "org_id_regex",
                "unwrap_list": True,
                "unquote": True,
                "pattern": r"\/(\w+:\w+)\/$",
                "optional": True,
            },
            "origin_id": {
                "data_type": "str_lower",
            },
        },
        "default_data_source": "OpenAhjoAPI",
    }

    tprek_config = {
        "has_meta": False,
        "next_key": None,
        "results_key": None,
        "fields": [
            "origin_id",
            "classification",
            "name",
            "parent",
        ],
        "update_fields": [
            "classification",
            "name",
            "parent",
        ],
        "field_config": {
            "parent": {
                "source_field": "parent_id",
                "data_type": "org_id",
                "optional": True,
            },
            "origin_id": {
                "source_field": "id",
                "data_type": "value",
            },
            "classification": {
                "source_field": "organization_type",
                "optional": True,
            },
            "name": {
                "source_field": "name_fi",
                "optional": True,
            },
        },
        "default_data_source": "tprek",
        "default_parent_organization": "Pääkaupunkiseudun toimipisterekisteri",
    }

    default_config = paatos_config
    timeout = 10

    def __init__(self, url, config=None):
        self.url = url

        self.config = copy.deepcopy(self.default_config)
        config = copy.deepcopy(config)
        if config:
            # only consider fields listed in given config when merging field configs
            default_field_config = self.config.pop("field_config")
            given_field_config = config.pop("field_config")
            merged_field_config = {}
            for field in config.get("fields", None):
                if field in given_field_config:
                    merged_field_config[field] = given_field_config[field]
                elif field in default_field_config:
                    merged_field_config[field] = default_field_config[field]
            self.config["field_config"] = merged_field_config
            self.config.update(config)
        logger.info(
            f"Importing organization data from {self.url} with the following config:"
        )
        logger.info(self.config)
        self.related_import_methods = {
            "data_source": self._import_data_source,
            "classification": self._import_organization_class,
            "parent": self._import_organization,
        }

        if self.config.get("deprecated", False):
            warnings.warn(
                "RestAPIImporter is initialized with a deprecated configuration.",
                DeprecationWarning,
                stacklevel=2,
            )

        self._organization_classes = {}
        self._data_sources = {}
        self._organizations = {}
        self._default_parent = None
        if self.config.get("default_parent_organization", None):
            origin_id_config = self.config["field_config"].get("origin_id", None)
            origin_id_field = (
                origin_id_config["source_field"] if origin_id_config else "origin_id"
            )
            data_source_config = self.config["field_config"].get("data_source", None)
            data_source_field = (
                data_source_config["source_field"]
                if data_source_config
                else "data_source"
            )
            name_config = self.config["field_config"].get("name", None)
            name_field = name_config["source_field"] if name_config else "name"
            default_parent_data = {
                origin_id_field: self.config["default_data_source"],
                data_source_field: self.config["default_data_source"],
                name_field: self.config["default_parent_organization"],
            }
            self._default_parent = self._import_organization(default_parent_data)
        # Save all the data from the REST endpoint in a single dict.
        # This is for cases where we need to access the data by id instead of link.
        # Caching is necessary for importing all parent organizations before
        # their children.
        self._data_dict = {}
        for data_item in self._data_iter(self.url):
            self._data_dict[data_item["id"]] = data_item

    @property
    def fields(self):
        return self.config["fields"]

    @property
    def update_fields(self):
        return self.config["update_fields"]

    @property
    def field_config(self):
        return self.config["field_config"]

    @property
    def has_meta(self):
        """Whether the response contains a separate object for pagination."""
        return self.config.get("has_meta", False)

    @property
    def next_key(self):
        """Object key that stores the link to the next page"""
        return self.config["next_key"]

    @property
    def results_key(self):
        """Object key that stores the list of organizations"""
        return self.config["results_key"]

    @property
    def rename_data_source(self):
        """Data source renaming configs"""
        return self.config.get("rename_data_source") or {}

    @property
    def skip_classifications(self):
        """List of organization classifications to skip"""
        return self.config.get("skip_classifications") or []

    @property
    def default_data_source(self):
        """Default data source string"""
        return self.config["default_data_source"]

    def _build_resource_url(self, resource_id) -> str:
        """Build an url for the given resource id.

        Resource identifiers in the data might be missing the scheme+host part.
        With
        - self.url = "https://dev.hel.fi/paatokset/v1/organization/" and
        - resource_id = "/paatokset/v1/organization/?limit=20&offset=20"

        the expected output is:
        "https://dev.hel.fi/paatokset/v1/organization/?limit=20&offset=20"
        """
        return urllib.parse.urljoin(self.url, resource_id)

    def _get_organization_class(self, data):
        """Get organization class for the given object data

        The method will first try to get the organization class from cache, and
        then get from database if not cached.
        """
        # organization class supports id, data_source, origin_id and name.
        supported_fields = {"id", "origin_id", "data_source", "name"}
        identifier = data.get("id")
        # organization class requires data source and origin_id.
        if isinstance(identifier, str) and ":" in identifier:
            if "data_source" not in data:
                data["data_source"] = identifier.split(":")[0]
            if "origin_id" not in data:
                data["origin_id"] = identifier.split(":")[1]
        # provided id used if origin id missing
        if "origin_id" not in data or not data["origin_id"]:
            data["origin_id"] = identifier
        # default data source used if missing
        if "data_source" not in data or not data["data_source"]:
            data["data_source"] = self.default_data_source
        data_source_instance = self.related_import_methods["data_source"](
            data["data_source"]
        )
        # reformat id to fit our model
        if not isinstance(identifier, str) or ":" not in identifier:
            data["id"] = data_source_instance.pk + ":" + str(identifier)
        data["data_source"] = data_source_instance
        # extra fields should not crash the import. Only use specified fields.
        data = {
            field: value for (field, value) in data.items() if field in supported_fields
        }
        if identifier not in self._organization_classes:
            defaults = {"name": data.pop("name", data["id"])}
            organization_class, _ = OrganizationClass.objects.get_or_create(
                **data, defaults=defaults
            )
            self._organization_classes[identifier] = organization_class
        return self._organization_classes[identifier]

    def _get_data_source(self, data):
        """Get data source for the given object data

        The method will first try to get the data source from cache, and
        then get from database if not cached.
        """
        identifier = data["id"]

        if identifier in self.rename_data_source:
            identifier = self.rename_data_source[identifier]
            data["id"] = identifier

        if identifier not in self._data_sources:
            data_source_model = get_data_source_model()
            data_source, _ = data_source_model.objects.get_or_create(**data)
            self._data_sources[identifier] = data_source
        return self._data_sources[identifier]

    def import_data(self):
        """Import data"""
        for data_item in self._data_dict.values():
            self._import_organization(data_item)

    @transaction.atomic
    def _import_organization(self, data):
        """Import single organization.

        The organization import is transactional so, for example, we do not
        accidentally save organization without parent if the parent organization
        saving failed for some reason.
        """
        # id must always be present
        config = self.field_config.get("origin_id") or {}
        if isinstance(data, dict):
            incoming_data = data
        else:
            raise DataImportError("Organization data must be dict.")
        origin_id = self._get_field_value(incoming_data, "origin_id", config)

        if origin_id in self._organizations:
            # No need to re-import/update an organization which is already
            # imported once in this run.
            logger.debug(f"Using cached Organization: {origin_id}")
            return self._organizations[origin_id]

        # data source may be missing altogether, or it may be optional
        data_source = None
        config = self.field_config.get("data_source") or {}
        try:
            data_source = self._get_field_value(incoming_data, "data_source", config)
        except DataImportError:
            if not config or config.get("optional"):
                pass
            else:
                raise
        if not data_source:
            data_source = self._get_field_value(
                {"data_source": self.default_data_source},
                "data_source",
                {"data_type": "value"},
            )

        logger.debug(f"Importing Organization: {origin_id}")

        # id is never imported in default config

        try:
            return self._import_organization_update(
                incoming_data, origin_id, data_source
            )
        except Organization.DoesNotExist:
            return self._import_organization_create(
                incoming_data, origin_id, data_source
            )

    def _import_organization_update(self, incoming_data, origin_id, data_source):
        """Update an existing organization upon import."""
        values_to_update = {}
        for field in self.update_fields:
            config = self.field_config.get(field) or {}
            try:
                value = self._get_field_value(incoming_data, field, config)
            except DataImportError:
                if config.get("optional"):
                    continue
                else:
                    raise
            values_to_update[field] = value

        # Organization parent (and its parent) have been imported and updated
        # recursively. If one of them used to be the descendant of this organization,
        # this might result in mptt InvalidMove exception if this organization instance
        # is not up-to-date with the database.
        # See https://github.com/django-mptt/django-mptt/issues/650

        # Therefore, we may only fetch the organization from the db *after* all its
        # parents have been processed, to get up-to-date status of the mptt tree before
        # saving each organization. Enforce lower case id standard, but recognize upper
        # case ids as equal:
        organization = Organization.objects.get(
            origin_id__iexact=origin_id, data_source=data_source
        )
        logger.info(f"Updating organization: {origin_id}")
        for field, value in values_to_update.items():
            setattr(organization, field, value)
        if (
            self.config.get("default_parent_organization", None)
            and not organization.parent
            and organization != self._default_parent
        ):
            organization.parent = self._default_parent
        organization.save()
        logger.debug(f"Caching Organization: {origin_id}")
        self._organizations[organization.origin_id] = organization

        return organization

    def _import_organization_create(self, incoming_data, origin_id, data_source):
        """Create a new Organization instance upon import."""
        object_data = {"origin_id": origin_id, "data_source": data_source}
        for field in self.fields:
            config = self.field_config.get(field) or {}
            try:
                object_data[field] = self._get_field_value(incoming_data, field, config)
            except DataImportError:
                if config.get("optional"):
                    continue
                else:
                    raise

        # Check if we're supposed to skip importing organizations of this class
        classification = object_data.get("classification")
        if classification and classification.origin_id in self.skip_classifications:
            self._organizations[origin_id] = None
            logger.info(
                f"Skipping organization: {origin_id} with type: {classification}"
            )
            return None

        logger.info(f"Creating Organization: {origin_id}")
        organization = Organization.objects.create(**object_data)
        if (
            self.config.get("default_parent_organization", None)
            and not organization.parent
            and organization != self._default_parent
        ):
            organization.parent = self._default_parent
            organization.save()
        logger.debug(f"Caching Organization: {origin_id}")
        self._organizations[origin_id] = organization

        return organization

    def _import_data_source(self, data):
        """Import data source

        If a single value is provided to data_source field, it assume it's
        the id of the data source.
        """
        if isinstance(data, dict):
            object_data = data
        else:
            object_data = {"id": data}

        return self._get_data_source(object_data)

    def _import_organization_class(self, data):
        """Import organization class.

        If a single value is provided to classification field, it assume it's
        the id of the organization class.

        If a dict is provided, origin_id is used if present, otherwise id.
        """
        if isinstance(data, dict):
            object_data = {k: v for k, v in data.items() if k != "id"}
            object_data["id"] = data["origin_id"] if "origin_id" in data else data["id"]
        else:
            object_data = {"id": data}

        return self._get_organization_class(object_data)

    def _data_iter(self, url):
        """Iterate over data items in the REST API endpoint.

        The iterator will follow over next page links if available.
        """
        logger.info(f"Start reading data from {url} ...")

        r = requests.get(url, timeout=self.timeout)
        try:
            r.raise_for_status()
        except requests.HTTPError as e:
            raise DataImportError(e) from e

        data = r.json()
        for data_item in data[self.results_key] if self.results_key else data:
            yield data_item

        logger.info(f"Reading data from {url} completed")

        if self.next_key:
            if self.has_meta:
                next_page_url = data["meta"][self.next_key]
            else:
                next_page_url = data[self.next_key]

            if next_page_url:
                yield from self._data_iter(self._build_resource_url(next_page_url))

    def _get_field_value(self, data_item, field, config):
        """Get source value for the field from the data item

        If source_field is specified in field config, the source_field
        will be used to get the source value, otherwise the model field
        will be used.

        If the data type is DataType.VALUE, the value will be returned;
        If the data type is DataType.STR_LOWER, the value will be returned stringified
        to lower case;
        If the data type is DataType.ORG_ID, corresponding organization will be
        returned;
        If the data type is DataType.LINK, the data in the link will be returned;
        If the data type is DataType.REGEX, the extracted value for the pattern will be
        returned.
        """
        source_field = config.get("source_field") or field
        try:
            value = data_item[source_field]
        except KeyError as e:
            raise DataImportError(
                f"Field not found in source data: {source_field}"
            ) from e

        if config.get("unwrap_list") and not value and isinstance(value, list):
            # Unwrap an empty list to avoid possible errors.
            value = None

        if not value:
            return value

        if config.get("data_type"):
            try:
                data_type = DataType(config["data_type"])
            except ValueError:
                raise DataImportError(
                    f"Invalid data type: {config['data_type']}. "
                    f"Supported data types are: "
                    f"{', '.join([e.value for e in DataType])}"
                )
        else:
            data_type = DataType.VALUE

        if data_type in (DataType.REGEX, DataType.ORG_ID_REGEX) and not config.get(
            "pattern"
        ):
            raise DataImportError(f"No regex pattern provided for the field: {field}")

        if config.get("unwrap_list"):
            if len(value) > 1:
                warnings.warn(
                    "More than one value in the list, unwrap_list only "
                    "unwraps the first value in the list.",
                    UserWarning,
                    stacklevel=2,
                )
            value = value[0]
        if config.get("unquote"):
            value = urllib.parse.unquote(value)

        if data_type == DataType.STR_LOWER:
            value = str(value).lower()
        elif data_type == DataType.ORG_ID:
            value = self._data_dict.get(value, None)
        elif data_type == DataType.ORG_ID_REGEX:
            value = self._get_regex_data(value, config["pattern"])
            value = self._data_dict.get(value, None)
        elif data_type == DataType.LINK:
            value = self._get_link_data(value)
        elif data_type == DataType.REGEX:
            value = self._get_regex_data(value, config["pattern"])

        # import related objects
        if field in self.related_import_methods:
            import_method = self.related_import_methods[field]
            value = import_method(value)

        return value

    @staticmethod
    def _get_regex_data(value, pattern):
        """Extract value from original string value with the given regex pattern"""
        match = re.search(pattern, value)
        if match:
            data = match.group(1)
        else:
            raise DataImportError(
                f"Cannot extract value from string {value} with pattern {pattern}"
            )
        return data

    def _get_link_data(self, value):
        """Get data fetched from the link"""
        validator = URLValidator()
        try:
            validator(value)
        except ValidationError as e:
            raise DataImportError(f"Invalid URL: {value}") from e

        r = requests.get(value, timeout=self.timeout)

        try:
            r.raise_for_status()
        except requests.HTTPError as e:
            raise DataImportError(e) from e

        return r.json()



# Generated by Django 1.11.7 on 2017-12-15 09:14
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0003_rm_responsible_org_add_internal_type"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="organization",
            options={
                "permissions": (
                    ("add_affiliated_organization", "Can add affiliated organization"),
                    (
                        "change_affiliated_organization",
                        "Can change affiliated organization",
                    ),
                    (
                        "delete_affiliated_organization",
                        "Can delete affiliated organization",
                    ),
                )
            },
        ),
    ]



# Generated by Django 1.11.7 on 2017-11-28 13:03
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="organization",
            name="replaced_by",
            field=models.OneToOneField(
                blank=True,
                help_text="The organization that replaces this organization",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="replaced_organization",
                to="django_orghierarchy.Organization",
            ),
        ),
    ]



# Generated by Django 2.2.8 on 2020-01-08 08:42

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0006_add_user_editable_to_data_source"),
    ]

    operations = [
        migrations.AlterField(
            model_name="organization",
            name="level",
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name="organization",
            name="lft",
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name="organization",
            name="rght",
            field=models.PositiveIntegerField(editable=False),
        ),
    ]



import copy

from django.core.management.base import BaseCommand, CommandError

from django_orghierarchy.importers import DataImportError, RestAPIImporter


class Command(BaseCommand):
    help = "Import organization data from a REST API"

    def add_arguments(self, parser):
        parser.add_argument(
            "url",
            metavar="URL",
            help="REST API endpoint from where the organization data will be imported",
        )

        parser.add_argument(
            "-s",
            "--rename-data-source",
            dest="rename_data_source",
            nargs="+",
            default=["helsinki:ahjo"],
            help="Rename data sources. Renaming should be specified as "
            "<old_identifier>:<new_identifier>",
        )

        parser.add_argument(
            "-c",
            "--config",
            dest="config",
            default="paatos",
            choices=["paatos", "tprek", "openahjo"],
            help="REST API configuration to use when importing.",
        )

    def _parse_rename_data_source(self, value):
        try:
            old_name, new_name = value.split(":")
        except ValueError:
            raise CommandError(
                "Invalid data source renaming. Renaming should be specified as "
                "<old_identifier>:<new_identifier>"
            )

        return old_name.strip(), new_name.strip()

    def handle(self, *args, **options):
        url = options["url"]

        config = {}
        if options["config"]:
            config = copy.deepcopy(
                getattr(RestAPIImporter, "%s_config" % options["config"])
            )
        if options["rename_data_source"]:
            rename_data_source = {}
            for item in options["rename_data_source"]:
                old_name, new_name = self._parse_rename_data_source(item)
                rename_data_source[old_name] = new_name
            config["rename_data_source"] = rename_data_source

        try:
            importer = RestAPIImporter(url, config)
            importer.import_data()
        except DataImportError as e:
            self.stderr.write("Import failed: {0}".format(e))
        else:
            self.stdout.write("Import completed successfully")



# Generated by Django 1.11.7 on 2017-11-23 15:13
import django.db.models.deletion
import mptt.fields
import swapper
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [swapper.dependency("django_orghierarchy", "DataSource")]

    operations = [
        migrations.CreateModel(
            name="DataSource",
            fields=[
                (
                    "id",
                    models.CharField(max_length=100, primary_key=True, serialize=False),
                ),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "swappable": swapper.swappable_setting(
                    "django_orghierarchy", "DataSource"
                ),
            },
        ),
        migrations.CreateModel(
            name="Organization",
            fields=[
                (
                    "id",
                    models.CharField(
                        editable=False,
                        max_length=255,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("origin_id", models.CharField(max_length=255, unique=True)),
                (
                    "name",
                    models.CharField(
                        help_text="A primary name, e.g. a legally recognized name",
                        max_length=255,
                    ),
                ),
                (
                    "founding_date",
                    models.DateField(
                        blank=True, help_text="A date of founding", null=True
                    ),
                ),
                (
                    "dissolution_date",
                    models.DateField(
                        blank=True, help_text="A date of dissolution", null=True
                    ),
                ),
                (
                    "created_time",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The time at which the resource was created",
                    ),
                ),
                (
                    "last_modified_time",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The time at which the resource was updated",
                    ),
                ),
                ("lft", models.PositiveIntegerField(db_index=True, editable=False)),
                ("rght", models.PositiveIntegerField(db_index=True, editable=False)),
                ("tree_id", models.PositiveIntegerField(db_index=True, editable=False)),
                ("level", models.PositiveIntegerField(db_index=True, editable=False)),
            ],
        ),
        migrations.CreateModel(
            name="OrganizationClass",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
            ],
            options={
                "verbose_name_plural": "Organization classes",
                "verbose_name": "Organization class",
            },
        ),
        migrations.AddField(
            model_name="organization",
            name="classification",
            field=models.ForeignKey(
                blank=True,
                help_text="An organization category, e.g. committee",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="django_orghierarchy.OrganizationClass",
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="created_by",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="created_organizations",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="data_source",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=swapper.get_model_name("django_orghierarchy", "DataSource"),
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="last_modified_by",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="modified_organizations",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="parent",
            field=mptt.fields.TreeForeignKey(
                blank=True,
                help_text="The organizations that contain this organization",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children",
                to="django_orghierarchy.Organization",
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="responsible_organization",
            field=models.ForeignKey(
                blank=True,
                help_text="Responsible organization",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="affiliated_organization",
                to="django_orghierarchy.Organization",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="organization",
            unique_together=set([("data_source", "origin_id")]),
        ),
        migrations.AddField(
            model_name="organization",
            name="admin_users",
            field=models.ManyToManyField(
                blank=True,
                related_name="admin_organizations",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="regular_users",
            field=models.ManyToManyField(
                blank=True,
                related_name="organization_memberships",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



# Generated by Django 3.1.2 on 2021-03-17 15:38

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        (
            "django_orghierarchy",
            "0009_add_change_organization_regular_users_permission",
        ),
    ]

    operations = [
        migrations.RenameField(
            model_name="datasource",
            old_name="user_editable",
            new_name="user_editable_organizations",
        ),
    ]



from django.contrib import admin
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import Permission
from django.test import RequestFactory, TestCase

from django_orghierarchy.admin import (
    AddAffiliatedOrganizationInline,
    AddSubOrganizationInline,
    AffiliatedOrganizationInline,
    OrganizationAdmin,
    ProtectedAffiliatedOrganizationInline,
    ProtectedSubOrganizationInline,
    SubOrganizationInline,
)
from django_orghierarchy.models import DataSource, Organization

from .factories import DataSourceFactory, OrganizationClassFactory, OrganizationFactory
from .utils import clear_user_perm_cache, make_admin


class TestDataSourceAdmin(TestCase):
    def test_data_source_admin_is_registered(self):
        is_registered = admin.site.is_registered(DataSource)
        self.assertTrue(is_registered)


class TestSubOrganizationInline(TestCase):
    def setUp(self):
        self.admin = make_admin()
        self.normal_admin = make_admin(username="normal_admin", is_superuser=False)

        self.site = AdminSite()
        self.factory = RequestFactory()

        self.normal_org = OrganizationFactory(internal_type=Organization.NORMAL)
        self.affiliated_org = OrganizationFactory(internal_type=Organization.AFFILIATED)
        self.editable_org = OrganizationFactory(
            data_source=(DataSourceFactory(user_editable_organizations=True))
        )

    def test_get_queryset(self):
        sub_org_inline = SubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        qs = sub_org_inline.get_queryset(request)
        self.assertQuerySetEqual(qs, [self.editable_org])

    def test_has_add_permission(self):
        sub_org_inline = SubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = sub_org_inline.has_add_permission(request, self.editable_org)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="add_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = sub_org_inline.has_add_permission(request, self.editable_org)
        self.assertFalse(has_perm)

    def test_has_change_permission(self):
        sub_org_inline = SubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = sub_org_inline.has_change_permission(request)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="change_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = sub_org_inline.has_change_permission(request)
        self.assertTrue(has_perm)

    def test_has_delete_permission(self):
        sub_org_inline = SubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = sub_org_inline.has_delete_permission(request)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="delete_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = sub_org_inline.has_delete_permission(request)
        self.assertTrue(has_perm)

    def test_readonly_fields(self):
        sub_org_inline = SubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        self.assertEqual(
            ("data_source", "origin_id", "id"),
            sub_org_inline.get_readonly_fields(request),
        )
        self.assertEqual(
            ("data_source", "origin_id", "id"),
            sub_org_inline.get_readonly_fields(request, obj=self.editable_org),
        )


class TestAddSubOrganizationInline(TestCase):
    def setUp(self):
        self.admin = make_admin()
        self.normal_admin = make_admin(username="normal_admin", is_superuser=False)

        self.site = AdminSite()
        self.factory = RequestFactory()

        self.editable_org = OrganizationFactory(
            data_source=(DataSourceFactory(user_editable_organizations=True))
        )

    def test_get_queryset(self):
        sub_org_inline = AddSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        qs = sub_org_inline.get_queryset(request)
        self.assertQuerySetEqual(qs, [])

    def test_has_add_permission(self):
        sub_org_inline = AddSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = sub_org_inline.has_add_permission(request, self.editable_org)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="add_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = sub_org_inline.has_add_permission(request, self.editable_org)
        self.assertTrue(has_perm)

    def test_has_change_permission(self):
        sub_org_inline = AddSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = sub_org_inline.has_change_permission(request)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="change_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = sub_org_inline.has_change_permission(request)
        self.assertFalse(has_perm)

    def test_has_delete_permission(self):
        sub_org_inline = AddSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = sub_org_inline.has_delete_permission(request)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="delete_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = sub_org_inline.has_delete_permission(request)
        self.assertFalse(has_perm)

    def test_readonly_fields(self):
        sub_org_inline = AddSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        self.assertEqual(("id",), sub_org_inline.get_readonly_fields(request))
        self.assertEqual(
            ("id",), sub_org_inline.get_readonly_fields(request, obj=self.editable_org)
        )


class TestProtectedSubOrganizationInline(TestCase):
    def setUp(self):
        self.admin = make_admin()
        self.site = AdminSite()
        self.factory = RequestFactory()

        self.normal_org = OrganizationFactory(internal_type=Organization.NORMAL)
        self.affiliated_org = OrganizationFactory(internal_type=Organization.AFFILIATED)
        self.editable_org = OrganizationFactory(
            data_source=(DataSourceFactory(user_editable_organizations=True))
        )

    def test_get_queryset(self):
        sub_org_inline = ProtectedSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        qs = sub_org_inline.get_queryset(request)
        self.assertQuerySetEqual(qs, [self.normal_org])

    def test_has_add_permission(self):
        sub_org_inline = ProtectedSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        has_perm = sub_org_inline.has_add_permission(request, self.editable_org)
        self.assertFalse(has_perm)

    def test_has_change_permission(self):
        sub_org_inline = ProtectedSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        has_perm = sub_org_inline.has_change_permission(request)
        # permission refers to the *parent* organization, change permission must be given to allow listing
        self.assertTrue(has_perm)

    def test_has_delete_permission(self):
        sub_org_inline = ProtectedSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        has_perm = sub_org_inline.has_delete_permission(request)
        self.assertFalse(has_perm)

    def test_readonly_fields(self):
        sub_org_inline = ProtectedSubOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        self.assertEqual(
            sub_org_inline.form.base_fields, sub_org_inline.get_readonly_fields(request)
        )


class TestAffiliatedOrganizationInline(TestCase):
    def setUp(self):
        self.admin = make_admin()
        self.normal_admin = make_admin(username="normal_admin", is_superuser=False)

        self.site = AdminSite()
        self.factory = RequestFactory()

        self.normal_org = OrganizationFactory(internal_type=Organization.NORMAL)
        self.affiliated_org = OrganizationFactory(internal_type=Organization.AFFILIATED)
        self.editable_org = OrganizationFactory(
            internal_type=Organization.AFFILIATED,
            data_source=(DataSourceFactory(user_editable_organizations=True)),
        )

    def test_get_queryset(self):
        aff_org_inline = AffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        qs = aff_org_inline.get_queryset(request)
        self.assertQuerySetEqual(qs, [self.editable_org])

    def test_has_add_permission(self):
        aff_org_inline = AffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = aff_org_inline.has_add_permission(request, self.editable_org)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="add_affiliated_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = aff_org_inline.has_add_permission(request, self.editable_org)
        self.assertFalse(has_perm)

    def test_has_change_permission(self):
        aff_org_inline = AffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = aff_org_inline.has_change_permission(request)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="change_affiliated_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = aff_org_inline.has_change_permission(request)
        self.assertTrue(has_perm)

    def test_has_delete_permission(self):
        aff_org_inline = AffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = aff_org_inline.has_delete_permission(request)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="delete_affiliated_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = aff_org_inline.has_delete_permission(request)
        self.assertTrue(has_perm)

    def test_readonly_fields(self):
        aff_org_inline = AffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        self.assertEqual(
            ("data_source", "origin_id", "id"),
            aff_org_inline.get_readonly_fields(request),
        )
        self.assertEqual(
            ("data_source", "origin_id", "id"),
            aff_org_inline.get_readonly_fields(request, obj=self.editable_org),
        )


class TestAddAffiliatedOrganizationInline(TestCase):
    def setUp(self):
        self.admin = make_admin()
        self.normal_admin = make_admin(username="normal_admin", is_superuser=False)

        self.site = AdminSite()
        self.factory = RequestFactory()

        self.editable_org = OrganizationFactory(
            data_source=(DataSourceFactory(user_editable_organizations=True))
        )

    def test_get_queryset(self):
        sub_org_inline = AddAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        qs = sub_org_inline.get_queryset(request)
        self.assertQuerySetEqual(qs, [])

    def test_has_add_permission(self):
        sub_org_inline = AddAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = sub_org_inline.has_add_permission(request, self.editable_org)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="add_affiliated_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = sub_org_inline.has_add_permission(request, self.editable_org)
        self.assertTrue(has_perm)

    def test_has_change_permission(self):
        sub_org_inline = AddAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = sub_org_inline.has_change_permission(request, self.editable_org)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="change_affiliated_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = sub_org_inline.has_change_permission(request)
        self.assertFalse(has_perm)

    def test_has_delete_permission(self):
        sub_org_inline = AddAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.normal_admin

        has_perm = sub_org_inline.has_delete_permission(request)
        self.assertFalse(has_perm)

        clear_user_perm_cache(self.normal_admin)
        perm = Permission.objects.get(codename="delete_affiliated_organization")
        self.normal_admin.user_permissions.add(perm)

        has_perm = sub_org_inline.has_delete_permission(request)
        self.assertFalse(has_perm)

    def test_readonly_fields(self):
        sub_org_inline = AddAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        self.assertEqual(("id",), sub_org_inline.get_readonly_fields(request))
        self.assertEqual(
            ("id",), sub_org_inline.get_readonly_fields(request, obj=self.editable_org)
        )


class TestProtectedAffiliatedOrganizationInline(TestCase):
    def setUp(self):
        self.admin = make_admin()
        self.site = AdminSite()
        self.factory = RequestFactory()

        self.normal_org = OrganizationFactory(internal_type=Organization.NORMAL)
        self.affiliated_org = OrganizationFactory(internal_type=Organization.AFFILIATED)
        self.editable_org = OrganizationFactory(
            internal_type=Organization.AFFILIATED,
            data_source=(DataSourceFactory(user_editable_organizations=True)),
        )

    def test_get_queryset(self):
        aff_org_inline = ProtectedAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        qs = aff_org_inline.get_queryset(request)
        self.assertQuerySetEqual(qs, [self.affiliated_org])

    def test_has_add_permission(self):
        aff_org_inline = ProtectedAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        has_perm = aff_org_inline.has_add_permission(request, self.editable_org)
        self.assertFalse(has_perm)

    def test_has_change_permission(self):
        aff_org_inline = ProtectedAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        has_perm = aff_org_inline.has_change_permission(request)
        # permission refers to the *parent* organization, change permission must be given to allow listing
        self.assertTrue(has_perm)

    def test_has_delete_permission(self):
        aff_org_inline = ProtectedAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        has_perm = aff_org_inline.has_delete_permission(request)
        self.assertFalse(has_perm)

    def test_readonly_fields(self):
        aff_org_inline = ProtectedAffiliatedOrganizationInline(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        self.assertEqual(
            aff_org_inline.form.base_fields, aff_org_inline.get_readonly_fields(request)
        )


class TestOrganizationAdmin(TestCase):
    def setUp(self):
        self.admin = make_admin()
        self.site = AdminSite()
        self.factory = RequestFactory()

        self.organization = OrganizationFactory()
        self.affiliated_organization = OrganizationFactory(
            internal_type=Organization.AFFILIATED, parent=self.organization
        )
        self.editable_organization = OrganizationFactory(
            data_source=(DataSourceFactory(user_editable_organizations=True))
        )

    def test_get_queryset(self):
        org = OrganizationFactory()
        sub_org = OrganizationFactory()
        another_sub_org = OrganizationFactory()
        sub_org.parent = self.organization
        another_sub_org.parent = org
        sub_org.save()
        another_sub_org.save()

        normal_admin = make_admin(username="normal_admin", is_superuser=False)

        oa = OrganizationAdmin(Organization, self.site)
        request = self.factory.get("/fake-url/")

        # test against superuser admin
        request.user = self.admin
        qs = oa.get_queryset(request)
        self.assertIsInstance(qs, Organization.objects._queryset_class)
        self.assertQuerySetEqual(
            qs,
            [
                self.organization,
                self.affiliated_organization,
                self.editable_organization,
                org,
                sub_org,
                another_sub_org,
            ],
            ordered=False,
        )

        # test against non-superuser admin
        request.user = normal_admin
        qs = oa.get_queryset(request)
        self.assertIsInstance(qs, Organization.objects._queryset_class)
        self.assertQuerySetEqual(qs, [])

        self.organization.admin_users.add(normal_admin)
        qs = oa.get_queryset(request)
        self.assertIsInstance(qs, Organization.objects._queryset_class)
        self.assertQuerySetEqual(
            qs,
            [
                self.organization,
                self.affiliated_organization,
                sub_org,
            ],
        )

        org.admin_users.add(normal_admin)
        qs = oa.get_queryset(request)
        self.assertIsInstance(qs, Organization.objects._queryset_class)
        self.assertQuerySetEqual(
            qs,
            [
                self.organization,
                self.affiliated_organization,
                org,
                sub_org,
                another_sub_org,
            ],
            ordered=False,
        )

    def test_save_model(self):
        oa = OrganizationAdmin(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        data_source = DataSourceFactory()
        organization_class = OrganizationClassFactory(data_source=data_source)
        organization = OrganizationFactory.build(
            classification=organization_class, data_source=data_source
        )
        oa.save_model(request, organization, None, None)
        self.assertEqual(organization.created_by, self.admin)
        self.assertEqual(organization.last_modified_by, self.admin)

        another_admin = make_admin(username="another_admin")
        request.user = another_admin
        oa.save_model(request, organization, None, None)
        self.assertEqual(organization.created_by, self.admin)
        self.assertEqual(organization.last_modified_by, another_admin)

    def test_move_model(self):
        # now, catching this problem requires a four-tier hierarchy and moving next-to-lowest tier
        # back and forth.
        oa = OrganizationAdmin(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        data_source = DataSourceFactory()
        organization_class = OrganizationClassFactory(data_source=data_source)
        organization = OrganizationFactory.build(
            classification=organization_class, data_source=data_source
        )
        organization2 = OrganizationFactory.build(
            classification=organization_class, data_source=data_source
        )
        organization3 = OrganizationFactory.build(
            classification=organization_class, data_source=data_source
        )
        organization4 = OrganizationFactory.build(
            classification=organization_class, data_source=data_source
        )
        organization5 = OrganizationFactory.build(
            classification=organization_class, data_source=data_source
        )
        organization6 = OrganizationFactory.build(
            classification=organization_class, data_source=data_source
        )
        organization7 = OrganizationFactory.build(
            classification=organization_class, data_source=data_source
        )
        organization2.parent = organization
        organization3.parent = organization
        organization4.parent = organization2
        organization5.parent = organization2
        organization6.parent = organization3
        organization7.parent = organization4
        oa.save_model(request, organization, None, None)
        oa.save_model(request, organization2, None, None)
        oa.save_model(request, organization3, None, None)
        oa.save_model(request, organization4, None, None)
        oa.save_model(request, organization5, None, None)
        oa.save_model(request, organization6, None, None)
        oa.save_model(request, organization7, None, None)
        self.assertEqual(organization4.parent, organization2)

        organization4.parent = organization3
        oa.save_model(request, organization4, None, None)
        self.assertEqual(organization4.parent, organization3)

        organization4.parent = organization2
        oa.save_model(request, organization4, None, None)
        self.assertEqual(organization4.parent, organization2)

    def test_indented_title(self):
        oa = OrganizationAdmin(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = self.admin

        self.assertNotIn("color: red;", oa.indented_title(self.organization))
        self.assertIn("color: red;", oa.indented_title(self.affiliated_organization))

    def test_has_change_permission(self):
        normal_admin = make_admin(username="normal_admin", is_superuser=False)

        oa = OrganizationAdmin(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = normal_admin

        has_perm = oa.has_change_permission(request)
        self.assertFalse(has_perm)

        clear_user_perm_cache(normal_admin)
        perm = Permission.objects.get(codename="change_affiliated_organization")
        normal_admin.user_permissions.add(perm)

        has_perm = oa.has_change_permission(request)
        self.assertTrue(has_perm)
        normal_admin.user_permissions.remove(perm)

        clear_user_perm_cache(normal_admin)
        perm = Permission.objects.get(codename="change_organization_regular_users")
        normal_admin.user_permissions.add(perm)

        has_perm = oa.has_change_permission(request)
        self.assertTrue(has_perm)

    def test_get_actions(self):
        normal_admin = make_admin(username="normal_admin", is_superuser=False)

        oa = OrganizationAdmin(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = normal_admin

        actions = oa.get_actions(request)
        self.assertNotIn("delete_selected", actions)

        clear_user_perm_cache(normal_admin)
        perm = Permission.objects.get(codename="delete_organization")
        normal_admin.user_permissions.add(perm)

        actions = oa.get_actions(request)
        self.assertIn("delete_selected", actions)

    def test_get_readonly_fields(self):
        normal_admin = make_admin(username="normal_admin", is_superuser=False)

        oa = OrganizationAdmin(Organization, self.site)
        request = self.factory.get("/fake-url/")
        request.user = normal_admin

        form_base_fields = OrganizationAdmin.form.base_fields
        oa_readonly_fields = OrganizationAdmin.readonly_fields
        oa_protected_readonly_fields = OrganizationAdmin.protected_readonly_fields

        fields = oa.get_readonly_fields(request)
        self.assertEqual(fields, oa_readonly_fields + ("replaced_by",))

        fields = oa.get_readonly_fields(request, self.organization)
        self.assertEqual(fields, form_base_fields)

        fields = oa.get_readonly_fields(request, self.affiliated_organization)
        self.assertEqual(fields, form_base_fields)

        fields = oa.get_readonly_fields(request, self.editable_organization)
        self.assertEqual(fields, form_base_fields)

        clear_user_perm_cache(normal_admin)
        perm = Permission.objects.get(codename="change_organization_regular_users")
        normal_admin.user_permissions.add(perm)

        fields = oa.get_readonly_fields(request)
        self.assertEqual(fields, oa_readonly_fields + ("replaced_by",))

        fields = oa.get_readonly_fields(request, self.organization)
        fields_minus_regular_users = list(tuple(form_base_fields))
        fields_minus_regular_users.remove("regular_users")
        self.assertEqual(list(tuple(fields)), fields_minus_regular_users)

        fields = oa.get_readonly_fields(request, self.affiliated_organization)
        self.assertEqual(list(tuple(fields)), fields_minus_regular_users)

        fields = oa.get_readonly_fields(request, self.editable_organization)
        self.assertEqual(list(tuple(fields)), fields_minus_regular_users)

        clear_user_perm_cache(normal_admin)
        perm = Permission.objects.get(codename="change_organization")
        normal_admin.user_permissions.add(perm)

        fields = oa.get_readonly_fields(request)
        self.assertEqual(fields, oa_readonly_fields + ("replaced_by",))

        fields = oa.get_readonly_fields(request, self.organization)
        self.assertEqual(fields, oa_protected_readonly_fields + ("replaced_by",))

        fields = oa.get_readonly_fields(request, self.affiliated_organization)
        self.assertEqual(fields, oa_protected_readonly_fields + ("replaced_by",))

        fields = oa.get_readonly_fields(request, self.editable_organization)
        self.assertEqual(
            fields,
            oa_readonly_fields
            + ("id", "data_source", "origin_id", "internal_type", "replaced_by"),
        )

        clear_user_perm_cache(normal_admin)
        perm = Permission.objects.get(codename="replace_organization")
        normal_admin.user_permissions.add(perm)

        fields = oa.get_readonly_fields(request)
        self.assertEqual(fields, oa_readonly_fields)

        fields = oa.get_readonly_fields(request, self.organization)
        self.assertEqual(fields, oa_protected_readonly_fields)

        fields = oa.get_readonly_fields(request, self.affiliated_organization)
        self.assertEqual(fields, oa_protected_readonly_fields)

        fields = oa.get_readonly_fields(request, self.editable_organization)
        self.assertEqual(
            fields,
            oa_readonly_fields + ("id", "data_source", "origin_id", "internal_type"),
        )



from django import forms
from django.forms import widgets
from django.utils.translation import gettext as _

from .models import Organization
from .utils import get_data_source_model


class OrganizationForm(forms.ModelForm):
    class Meta:
        model = Organization
        fields = (
            "data_source",
            "origin_id",
            "classification",
            "name",
            "founding_date",
            "dissolution_date",
            "internal_type",
            "parent",
            "admin_users",
            "regular_users",
            "replaced_by",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # the fields can be dynamically exclude, for example set them to readonly in
        # admin
        if "replaced_by" in self.fields:
            # An organization cannot be replaced by an organization that have already
            # been replaced
            qs = Organization.objects.filter(replaced_by__isnull=True)
            if self.instance.id:
                # prevent self reference
                qs = qs.exclude(id=self.instance.id)
            self.fields["replaced_by"].queryset = qs

        if "data_source" in self.fields:
            # Only allow selecting data source within editable sources
            self.fields[
                "data_source"
            ].queryset = get_data_source_model().objects.filter(
                user_editable_organizations=True
            )
        if "parent" in self.fields and self.instance.id:
            # prevent recursive reference
            desc_ids = self.instance.get_descendants(include_self=True).values_list(
                "id", flat=True
            )
            self.fields["parent"].queryset = Organization.objects.exclude(
                id__in=desc_ids
            )

    def clean(self):
        cleaned_data = super().clean()
        internal_type = cleaned_data.get("internal_type")
        parent = cleaned_data.get("parent")

        if internal_type == Organization.AFFILIATED and parent is None:
            raise forms.ValidationError(
                _("Affiliated organization must have a parent organization")
            )

        return cleaned_data


class SubOrganizationForm(forms.ModelForm):
    default_internal_type = Organization.NORMAL

    class Meta:
        model = Organization
        fields = (
            "name",
            "founding_date",
            "classification",
            "data_source",
            "origin_id",
            "internal_type",
        )
        widgets = {
            "internal_type": widgets.HiddenInput,
        }

    def __init__(self, *args, **kwargs):
        if "initial" not in kwargs:
            kwargs["initial"] = {}
        kwargs["initial"]["internal_type"] = self.default_internal_type
        super().__init__(*args, **kwargs)

        # the fields can be dynamically exclude, for example set them to readonly in
        # admin
        if "data_source" in self.fields:
            # Only allow selecting data source within editable sources
            self.fields[
                "data_source"
            ].queryset = get_data_source_model().objects.filter(
                user_editable_organizations=True
            )

    def clean_internal_type(self):
        return self.initial["internal_type"]  # do not allow changing internal_type


class AffiliatedOrganizationForm(SubOrganizationForm):
    default_internal_type = Organization.AFFILIATED



from django.test import TestCase

from django_orghierarchy.forms import (
    AffiliatedOrganizationForm,
    OrganizationForm,
    SubOrganizationForm,
)
from django_orghierarchy.models import Organization

from .factories import DataSourceFactory, OrganizationClassFactory, OrganizationFactory


class TestOrganizationForm(TestCase):
    def setUp(self):
        self.organization = OrganizationFactory()
        self.data_source = DataSourceFactory(user_editable_organizations=True)

    def test_init_without_instance(self):
        form = OrganizationForm()
        self.assertQuerySetEqual(form.fields["parent"].queryset, [self.organization])

    def test_init_with_instance(self):
        form = OrganizationForm(instance=self.organization)
        self.assertQuerySetEqual(form.fields["parent"].queryset, [])

    def test_replaced_by_field_queryset_exclude_already_replaced(self):
        OrganizationFactory(replaced_by=self.organization)
        organization_2 = OrganizationFactory()

        form = OrganizationForm()
        queryset = form.fields["replaced_by"].queryset
        self.assertQuerySetEqual(
            queryset, [self.organization, organization_2], ordered=False
        )

        form = OrganizationForm(instance=organization_2)
        queryset = form.fields["replaced_by"].queryset
        self.assertQuerySetEqual(queryset, [self.organization])

    def test_clean(self):
        form_data = {
            "data_source": self.data_source.id,
            "origin_id": "abc",
            "classification": self.organization.classification_id,
            "name": "test org",
            "internal_type": Organization.AFFILIATED,
            "parent": "",
        }
        form = OrganizationForm(form_data)
        self.assertFalse(form.is_valid())

        form_data["parent"] = self.organization.id
        form = OrganizationForm(form_data)
        self.assertTrue(form.is_valid())


class TestSubOrganizationForm(TestCase):
    def setUp(self):
        self.data_source = DataSourceFactory(user_editable_organizations=True)
        self.organization_class = OrganizationClassFactory()

    def test__init__(self):
        form = SubOrganizationForm()
        self.assertEqual(form.initial["internal_type"], Organization.NORMAL)

        form = SubOrganizationForm(initial={"internal_type": Organization.AFFILIATED})
        self.assertEqual(form.initial["internal_type"], Organization.NORMAL)

    def test_clean_internal_type(self):
        form_data = {
            "internal_type": Organization.NORMAL,
            "data_source": self.data_source.id,
            "origin_id": "sub-org",
            "classification": self.organization_class.id,
            "name": "sub organization",
        }
        form = SubOrganizationForm(form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["internal_type"], Organization.NORMAL)

        form_data["internal_type"] = Organization.AFFILIATED
        form = SubOrganizationForm(form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["internal_type"], Organization.NORMAL)


class TestAffiliatedOrganizationForm(TestCase):
    def setUp(self):
        self.data_source = DataSourceFactory(user_editable_organizations=True)
        self.organization_class = OrganizationClassFactory()

    def test__init__(self):
        form = AffiliatedOrganizationForm()
        self.assertEqual(form.initial["internal_type"], Organization.AFFILIATED)

        form = AffiliatedOrganizationForm(
            initial={"internal_type": Organization.NORMAL}
        )
        self.assertEqual(form.initial["internal_type"], Organization.AFFILIATED)

    def test_clean_internal_type(self):
        form_data = {
            "internal_type": Organization.NORMAL,
            "data_source": self.data_source.id,
            "origin_id": "sub-org",
            "classification": self.organization_class.id,
            "name": "sub organization",
        }
        form = AffiliatedOrganizationForm(form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["internal_type"], Organization.AFFILIATED)

        form_data["internal_type"] = Organization.AFFILIATED
        form = AffiliatedOrganizationForm(form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data["internal_type"], Organization.AFFILIATED)



# Generated by Django 2.2.8 on 2020-01-08 08:44

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0007_set_editable_false_on_mptt_fields"),
    ]

    operations = [
        migrations.AlterField(
            model_name="organization",
            name="created_by",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="created_organizations",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="organization",
            name="last_modified_by",
            field=models.ForeignKey(
                blank=True,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="modified_organizations",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="organization",
            name="replaced_by",
            field=models.OneToOneField(
                blank=True,
                help_text="The organization that replaces this organization",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="replaced_organization",
                to="django_orghierarchy.Organization",
            ),
        ),
    ]



# Generated by Django 1.11.10 on 2018-08-27 13:31
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0005_add_fields_to_organization_class"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="organization",
            options={
                "permissions": (
                    ("add_affiliated_organization", "Can add affiliated organization"),
                    (
                        "change_affiliated_organization",
                        "Can change affiliated organization",
                    ),
                    (
                        "delete_affiliated_organization",
                        "Can delete affiliated organization",
                    ),
                    (
                        "replace_organization",
                        "Can replace an organization with a new one",
                    ),
                )
            },
        ),
        migrations.AddField(
            model_name="datasource",
            name="user_editable",
            field=models.BooleanField(
                default=False, verbose_name="Objects may be edited by users"
            ),
        ),
    ]



import swapper
from django.conf import settings
from django.db import models
from django.db.models import Q, UniqueConstraint
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from mptt.models import MPTTModel, TreeForeignKey


class AbstractDataSource(models.Model):
    """Abstract data source model.

    Abstract data source model that provides required fields
    for custom data source model.
    """

    id = models.CharField(max_length=100, primary_key=True)
    name = models.CharField(max_length=255)
    user_editable_organizations = models.BooleanField(
        default=False, verbose_name=_("Organizations may be edited by users")
    )

    class Meta:
        abstract = True

    def __str__(self):
        return self.name


class DataSource(AbstractDataSource):
    """Default data source model.

    The default data source model will be used if the project
    does not provide a custom data source model.
    """

    class Meta:
        swappable = swapper.swappable_setting("django_orghierarchy", "DataSource")


class DataModel(models.Model):
    id = models.CharField(max_length=255, primary_key=True, editable=False)
    data_source = models.ForeignKey(
        swapper.get_model_name("django_orghierarchy", "DataSource"),
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    origin_id = models.CharField(max_length=255, blank=True)
    created_time = models.DateTimeField(
        default=timezone.now, help_text=_("The time at which the resource was created")
    )
    last_modified_time = models.DateTimeField(
        auto_now=True, help_text=_("The time at which the resource was updated")
    )

    class Meta:
        abstract = True
        constraints = [
            UniqueConstraint(
                fields=("data_source", "origin_id"),
                name="%(app_label)s_%(class)s_data_source_and_origin_id_unique",
            ),
            # PostgreSQL considers null as always unique
            # so origin_id must be unique among null data sources
            UniqueConstraint(
                fields=("origin_id",),
                condition=Q(data_source__isnull=True),
                name="%(app_label)s_%(class)s_origin_id_unique_if_data_source_null",
            ),
        ]

    def save(self, *args, **kwargs):
        if not self.id:
            # the id is only set when creating object, it cannot be changed later
            self.id = "{0}:{1}".format(self.data_source_id, self.origin_id)
            # Last resort in case unique validations don't catch it:
            # set force_insert to True to avoid accidentally overwriting
            # an existing row
            kwargs["force_insert"] = True
        super().save(*args, **kwargs)


class OrganizationClass(DataModel):
    name = models.CharField(max_length=255)

    class Meta:
        constraints = DataModel.Meta.constraints
        verbose_name = _("Organization class")
        verbose_name_plural = _("Organization classes")

    def __str__(self):
        return self.name


class Organization(MPTTModel, DataModel):
    NORMAL = "normal"
    AFFILIATED = "affiliated"

    INTERNAL_TYPES = (
        (NORMAL, _("Normal organization")),
        (AFFILIATED, _("Affiliated organization")),
    )

    internal_type = models.CharField(
        max_length=20, choices=INTERNAL_TYPES, default=NORMAL
    )

    classification = models.ForeignKey(
        OrganizationClass,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        help_text=_("An organization category, e.g. committee"),
    )
    name = models.CharField(
        max_length=255, help_text=_("A primary name, e.g. a legally recognized name")
    )
    founding_date = models.DateField(
        blank=True, null=True, help_text=_("A date of founding")
    )
    dissolution_date = models.DateField(
        blank=True, null=True, help_text=_("A date of dissolution")
    )
    parent = TreeForeignKey(
        "self",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        related_name="children",
        help_text=_("The organizations that contain this organization"),
    )
    admin_users = models.ManyToManyField(
        settings.AUTH_USER_MODEL, blank=True, related_name="admin_organizations"
    )
    regular_users = models.ManyToManyField(
        settings.AUTH_USER_MODEL, blank=True, related_name="organization_memberships"
    )
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        related_name="created_organizations",
        null=True,
        blank=True,
        editable=False,
    )
    last_modified_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        related_name="modified_organizations",
        null=True,
        blank=True,
        editable=False,
    )
    replaced_by = models.OneToOneField(
        "self",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="replaced_organization",
        help_text=_("The organization that replaces this organization"),
    )

    @cached_property
    def sub_organizations(self):
        return self.children.filter(internal_type=self.NORMAL)

    @cached_property
    def affiliated_organizations(self):
        return self.children.filter(internal_type=self.AFFILIATED)

    class Meta:
        constraints = DataModel.Meta.constraints
        permissions = (
            ("add_affiliated_organization", "Can add affiliated organization"),
            ("change_affiliated_organization", "Can change affiliated organization"),
            ("delete_affiliated_organization", "Can delete affiliated organization"),
            ("replace_organization", "Can replace an organization with a new one"),
            (
                "change_organization_regular_users",
                "Can add/remove regular users to organizations",
            ),
        )

    def __str__(self):
        if self.dissolution_date:
            return self.name + " (dissolved)"
        return self.name



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="django-orghierarchy",
    version="0.6.0",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    license="MIT License",
    description="Reusable Django application for hierarchical organizations.",
    long_description=README,
    url="https://github.com/City-of-Helsinki/django-orghierarchy",
    author="City of Helsinki",
    author_email="dev@hel.fi",
    python_requires=">=3.9",
    install_requires=[
        "Django>=4.2",
        "django-mptt",
        "djangorestframework",
        "requests",
        "swapper",
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.0",
        "Framework :: Django :: 5.1",
        "Framework :: Django :: 5.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



# Generated by Django 2.2.8 on 2020-02-03 12:04

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_orghierarchy", "0008_modify_organization_on_deletes"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="organization",
            options={
                "permissions": (
                    ("add_affiliated_organization", "Can add affiliated organization"),
                    (
                        "change_affiliated_organization",
                        "Can change affiliated organization",
                    ),
                    (
                        "delete_affiliated_organization",
                        "Can delete affiliated organization",
                    ),
                    (
                        "replace_organization",
                        "Can replace an organization with a new one",
                    ),
                    (
                        "change_organization_regular_users",
                        "Can add/remove regular users to organizations",
                    ),
                )
            },
        ),
    ]



from django.db import IntegrityError
from django.test import TestCase

from django_orghierarchy.models import Organization

from .factories import DataSourceFactory, OrganizationClassFactory, OrganizationFactory


class TestDataSource(TestCase):
    def setUp(self):
        self.data_source = DataSourceFactory(name="test name")

    def test__str__(self):
        self.assertEqual(self.data_source.__str__(), "test name")


class TestOrganizationClass(TestCase):
    def setUp(self):
        self.organization_class = OrganizationClassFactory(name="test name")

    def test__str__(self):
        self.assertEqual(self.organization_class.__str__(), "test name")


class TestOrganization(TestCase):
    def setUp(self):
        self.parent_organization = OrganizationFactory(name="parent name")
        self.organization = OrganizationFactory(
            name="test name", parent=self.parent_organization
        )
        self.affiliated_organization = OrganizationFactory(
            name="test aff org",
            parent=self.parent_organization,
            internal_type=Organization.AFFILIATED,
        )

    def test__str__(self):
        self.assertEqual(self.parent_organization.__str__(), "parent name")
        self.assertEqual(self.organization.__str__(), "test name")

    def test_save(self):
        data_source = DataSourceFactory(id="data-source")
        organization = OrganizationFactory(data_source=data_source, origin_id="ABC123")
        self.assertEqual(organization.id, "data-source:ABC123")

        organization.origin_id = "XYZ"
        organization.save()
        # test the id is not changed
        self.assertEqual(organization.id, "data-source:ABC123")

    def test_sub_organizations(self):
        qs = self.parent_organization.sub_organizations
        self.assertQuerySetEqual(qs, [self.organization])

    def test_affiliated_organizations(self):
        qs = self.parent_organization.affiliated_organizations
        self.assertQuerySetEqual(qs, [self.affiliated_organization])

    def test_save_duplicate_id_when_data_source_null(self):
        OrganizationFactory(data_source=None, origin_id="test")
        with self.assertRaises(IntegrityError):
            OrganizationFactory(data_source=None, origin_id="test")

    def test_save_duplicate(self):
        org1 = OrganizationFactory()
        with self.assertRaises(IntegrityError):
            OrganizationFactory(data_source=org1.data_source, origin_id=org1.origin_id)
